# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
import pytest


from tango import (
    Attribute,
    MultiAttrProp,
    DevEncoded,
)
from tango.server import Device
from tango.server import command, attribute
from tango.test_utils import (
    DeviceTestContext,
)


class SimpleDevice(Device):
    @attribute(dtype=bool)
    def attr_bool(self):
        pass

    @attribute(dtype=str)
    def attr_str(self):
        pass

    @attribute(dtype=DevEncoded)
    def attr_enc(self):
        pass

    def check_multi_attr(self, attr_name):
        attr = self.get_device_attr().get_attr_by_name(attr_name)
        assert isinstance(attr, Attribute)

        prop = attr.get_properties()
        assert isinstance(prop, MultiAttrProp)

        prop = MultiAttrProp()
        ret = attr.get_properties(prop)
        assert isinstance(prop, MultiAttrProp)
        assert prop is ret

        with pytest.raises(
            TypeError, match="attr_cfg must be an instance of MultiAttrProp"
        ):
            attr.get_properties([])

        assert prop.label == attr_name
        assert prop.unit == ""

        prop.unit = "kgm^2/s"

        attr.set_properties(prop)

        with pytest.raises(
            TypeError, match="attr_cfg must be an instance of MultiAttrProp"
        ):
            attr.set_properties([])

    @command
    def check_attr_str(self):
        self.check_multi_attr("attr_str")

    @command
    def check_attr_enc(self):
        self.check_multi_attr("attr_enc")

    @command
    def check_attr_bool(self):
        self.check_multi_attr("attr_bool")


def test_stuff():
    with DeviceTestContext(SimpleDevice) as proxy:
        proxy.check_attr_str()
        info = proxy.attribute_query("attr_str")
        assert info.unit == "kgm^2/s"

        proxy.check_attr_enc()
        info = proxy.attribute_query("attr_enc")
        assert info.unit == "kgm^2/s"

        proxy.check_attr_bool()
        info = proxy.attribute_query("attr_bool")
        assert info.unit == "kgm^2/s"



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
This is an internal PyTango module.
"""

__all__ = ("time_val_init",)

__docformat__ = "restructuredtext"

import time
import datetime

from tango import TimeVal
import numbers


def __TimeVal__init(self, a=None, b=None, c=None):
    TimeVal.__init_original(self)
    if a is None:
        return

    if isinstance(a, datetime.datetime):
        assert b is None and c is None
        a = time.mktime(a.timetuple()) + a.microsecond * 1e-6

    if isinstance(a, numbers.Number):
        if b is None:
            self.tv_sec = int(a)
            usec = (a - self.tv_sec) * 1e6
            self.tv_usec = int(usec)
            self.tv_nsec = int((usec - self.tv_usec) * 1e3)
        else:
            self.tv_sec, self.tv_usec, self.tv_nsec = a, b, c


def __TimeVal__totime(self):
    """
    totime(self) -> float

        Returns a float representing this time value

        Parameters : None
        Return     : a float representing the time value

    .. versionadded:: 7.1.0"""
    return self.tv_sec + 1e-6 * self.tv_usec + 1e-9 * self.tv_nsec


def __TimeVal__todatetime(self):
    """
    todatetime(self) -> datetime.datetime

        Returns a :class:`datetime.datetime` object representing
        the same time value

        Parameters : None
        Return     : (datetime.datetime) the time value in datetime format

    .. versionadded:: 7.1.0"""
    return datetime.datetime.fromtimestamp(self.totime())


def __TimeVal__fromtimestamp(ts):
    """
    fromtimestamp(ts) -> TimeVal

        A static method returning a :class:`tango.TimeVal` object representing
        the given timestamp

        Parameters :
            - ts : (float) a timestamp
        Return     : (TimeVal) representing the given timestamp

    .. versionadded:: 7.1.0"""
    return TimeVal(ts)


def __TimeVal__fromdatetime(dt):
    """
    fromdatetime(dt) -> TimeVal

        A static method returning a :class:`tango.TimeVal` object representing
        the given :class:`datetime.datetime`

        Parameters :
            - dt : (datetime.datetime) a datetime object
        Return     : (TimeVal) representing the given timestamp

    .. versionadded:: 7.1.0

    .. versionadded:: 7.1.2
        Documented
    """
    return TimeVal(dt)


def __TimeVal__now():
    """
    now() -> TimeVal

        A static method returning a :class:`tango.TimeVal` object representing
        the current time

        Parameters : None
        Return     : (TimeVal) representing the current time

    .. versionadded:: 7.1.0

    .. versionadded:: 7.1.2
        Documented
    """
    return TimeVal(time.time())


def __TimeVal__strftime(self, format):
    """
    strftime(self, format) -> str

        Convert a time value to a string according to a format specification.

        Parameters :
            format : (str) See the python library reference manual for formatting codes
        Return     : (str) a string representing the time according to a format specification.

    .. versionadded:: 7.1.0

    .. versionadded:: 7.1.2
        Documented
    """
    return self.todatetime().strftime(format)


def __TimeVal__isoformat(self, sep="T"):
    """
    isoformat(self, sep='T') -> str

        Returns a string in ISO 8601 format, YYYY-MM-DDTHH:MM:SS[.mmmmmm][+HH:MM]

        Parameters :
            sep : (str) sep is used to separate the year from the time, and defaults to 'T'
        Return     : (str) a string representing the time according to a format specification.

    .. versionadded:: 7.1.0

    .. versionadded:: 7.1.2
        Documented

    .. versionchanged:: 7.1.2
        The `sep` parameter is not mandatory anymore and defaults to 'T' (same as :meth:`datetime.datetime.isoformat`)
    """
    return self.todatetime().isoformat(sep)


def __TimeVal__str__(self):
    """
    __str__(self) -> str

        Returns a string representation of TimeVal

        Parameters : None
        Return     : (str) a string representing the time (same as :class:`datetime.datetime`)

    .. versionadded:: 7.1.0

    .. versionadded:: 7.1.2
        Documented
    """
    return str(self.todatetime())


def time_val_init():
    TimeVal.__init_original = TimeVal.__init__
    TimeVal.__init__ = __TimeVal__init
    TimeVal.totime = __TimeVal__totime
    TimeVal.todatetime = __TimeVal__todatetime
    TimeVal.fromtimestamp = staticmethod(__TimeVal__fromtimestamp)
    TimeVal.fromdatetime = staticmethod(__TimeVal__fromdatetime)
    TimeVal.now = staticmethod(__TimeVal__now)
    TimeVal.strftime = __TimeVal__strftime
    TimeVal.isoformat = __TimeVal__isoformat
    TimeVal.__str__ = __TimeVal__str__



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
import numpy as np
import pytest

from tango import Database
from tango.server import Device, attribute
from tango.test_utils import MultiDeviceTestContext, assert_close


class DeviceToTest(Device):
    @attribute(dtype=str)
    def attr1(self):
        return "attr1"

    @attribute(dtype=str)
    def attr2(self):
        return "attr2"


device_name = "device/test/1"

devices_info = ({"class": DeviceToTest, "devices": [{"name": device_name}]},)


@pytest.fixture()
def test_database():
    with MultiDeviceTestContext(devices_info) as context:
        yield Database(context.db)


def test_put_remove_attribute_properties(test_database):
    attr1_properties = {"attr1": {"value1": ["1"], "value2": ["2"], "value3": ["3"]}}
    attr2_properties = {"attr2": {"value1": ["4"], "value2": ["5"], "value3": ["6"]}}

    test_database.put_device_attribute_property(
        device_name, attr1_properties | attr2_properties
    )

    assert_close(
        test_database.get_device_attribute_property(device_name, "attr1"),
        attr1_properties,
    )
    assert_close(
        test_database.get_device_attribute_property(device_name, "attr2"),
        attr2_properties,
    )

    attr1_prop_to_delete = {"attr1": ["value1"]}
    attr2_prop_to_delete = {"attr2": ["value2"]}

    test_database.delete_device_attribute_property(
        device_name, attr1_prop_to_delete | attr2_prop_to_delete
    )

    [attr1_properties["attr1"].pop(k, None) for k in attr1_prop_to_delete["attr1"]]
    [attr2_properties["attr2"].pop(k, None) for k in attr2_prop_to_delete["attr2"]]

    assert_close(
        test_database.get_device_attribute_property(device_name, "attr1"),
        attr1_properties,
    )
    assert_close(
        test_database.get_device_attribute_property(device_name, "attr2"),
        attr2_properties,
    )


class BadReprObj:
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return f"don't use the repr! {self.value}"


def test_put_get_attribute_properties_string_conversion(test_database):
    attr_properties = {
        "attr": {
            "value0": ["zero"],
            "value1": ["o", "n", "e"],
            "value2": [2],
            "value3": [3.0],
            "value4": [np.float64(4.0)],
            "value5": [BadReprObj(5)],
            "value6": 6,
            "value7": 7.0,
            "value8": np.float64(8.0),
            "value9": BadReprObj(9),
            "value10": True,
        }
    }
    expected_string_properties = {
        "attr": {
            "value0": ["zero"],
            "value1": ["o", "n", "e"],
            "value2": ["2"],
            "value3": ["3.0"],
            "value4": ["4.0"],
            "value5": ["5"],
            "value6": ["6"],
            "value7": ["7.0"],
            "value8": ["8.0"],
            "value9": ["9"],
            "value10": ["True"],
        }
    }

    test_database.put_device_attribute_property(device_name, attr_properties)

    assert_close(
        test_database.get_device_attribute_property(device_name, "attr"),
        expected_string_properties,
    )



#!/usr/bin/env python3
# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
Simple power supply "hardware" simulator, communicating with plain
text commands over TCP.  For uses with the ps1*.py and ps2*.py device
servers.

Requires gevent to be installed.
"""

import time
import logging
import gevent.server

DEFAULT_BIND = ""
DEFAULT_PORT = 45000


class Attr:
    def __init__(
        self,
        *,
        initial_value=0.0,
        encode=lambda x: bytes(str(x), "ascii"),
        decode=float,
    ):
        self.value = initial_value
        self.encode = encode
        self.decode = decode

    def get(self):
        return self.encode(self.value)

    def set(self, value):
        self.value = self.decode(value)


class Calibrate(Attr):
    def set(self, value):
        self.ts = time.time()
        super().set(value)


class State(Attr):
    def __init__(self, calib, *args, **kwargs):
        kwargs["initial_value"] = 0
        kwargs["decode"] = int
        super().__init__(*args, **kwargs)
        self.calib = calib
        calib.ts = 0

    def get(self):
        self.value = 0
        if time.time() - self.calib.ts < 2:
            self.value = 1
        return super().get()


class PSSimulator(gevent.server.StreamServer):
    class Error(Exception):
        pass

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.log = logging.getLogger(f"simulator.{self.server_port}")
        calib = Calibrate(initial_value=0)
        self.attrs = {
            b"stat": State(calib),
            b"vol": Attr(initial_value=0.1),
            b"curr": Attr(initial_value=0.0),
            b"calib": calib,
        }

    def __getitem__(self, name):
        return self.attrs[name].get()

    def __setitem__(self, name, value):
        self.attrs[name].set(value)

    def handle(self, sock, addr):
        log = self.log
        log.info("new connection from %r", addr)
        fileobj = sock.makefile(mode="rb")
        while True:
            request = fileobj.readline()
            if not request:
                log.info("disconnected %r", addr)
                break
            log.info("request %r", request)
            reply = b"ERROR"
            try:
                reply = self.handle_request(request)
            except PSSimulator.Error:
                pass
            except Exception:
                log.exception("Unforeseen error")
            gevent.sleep(1e-1)
            sock.sendall(reply + b"\n")
            log.info("replyed %r", reply)
        fileobj.close()

    def handle_request(self, request):
        req_lower = request.strip().lower()
        is_query = b"?" in req_lower
        pars = req_lower.split()
        name = pars[0]
        if is_query:
            name = name[:-1]  # take out '?'
        if is_query:
            return self[name]
        self[name] = pars[1]
        return b"OK"


def main(number=1, bind=DEFAULT_BIND, port=DEFAULT_PORT, **kwargs):
    servers = []
    logging.info("starting simulator...")
    for i in range(number):
        address = bind, port + i
        server = PSSimulator(address)
        server.start()
        servers.append(server)
        server.log.info("simulator listenning on %r!", address)
    try:
        while True:
            gevent.sleep(1)
    except KeyboardInterrupt:
        logging.info("Ctrl-C pressed. Bailing out!")
        for server in servers:
            server.stop()


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--port", type=int, default=DEFAULT_PORT)
    parser.add_argument("--bind", default=DEFAULT_BIND)
    parser.add_argument("--log-level", default="info")
    parser.add_argument("--number", type=int, default=1)
    args = parser.parse_args()
    logging.basicConfig(level=args.log_level.upper())
    main(**vars(args))



#!/usr/bin/env python3
# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
#
# client example: executing commands

import sys
from tango import DeviceProxy, DevFailed, Except

dev_name = "sys/tg_test/1"
cmd_name = "DevShort"

# step 1 create the device proxy
try:
    dev = DeviceProxy(dev_name)
    print(f"proxy for {dev_name} created")
except DevFailed as ex:
    print("failed to create DeviceProxy")
    Except.print_exception(ex)
    sys.exit(1)

try:
    data_out = dev.command_inout("DevShort", 1234)
    print(f"command_inout DevShort result: {data_out}")
    my_dbl = 1.22  # here my_dbl is a float
    my_str = "ABC"  # here my_str is a string
    print(
        f"types of variables before data_out extraction: "
        f"{type(data_out)}, {type(my_dbl)}, {type(my_str)}"
    )
    my_dbl = data_out  # here my_dbl turns into a short
    my_str = data_out  # here my_str turns into a short
    print(
        f"command_inout DevVoid results extracted to existing variable: "
        f"{data_out}, {my_dbl}, {my_str}, {type(data_out)}, {type(my_dbl)}, {type(my_str)}"
    )
    # type of variables changes if handled by python! powerful, easy, can play tricks...
except DevFailed as ex:
    print("failed to execute command DevShort")

# now with DevVoid
try:
    dev.command_inout("DevVoid")
    print("command_inout DevVoid OK")
except DevFailed as ex:
    print("failed to execute command DevVoid")

"""
Typical output:

➜  training git:(develop) ✗ docker-compose exec cli /training/client/command01.py
proxy for sys/tg_test/1 created
command_inout DevShort result: 1234
types of variables before data_out extraction: <class 'int'>, <class 'float'>, <class 'str'>
command_inout DevVoid results extracted to existing variable: 1234, 1234, 1234, <class 'int'>, <class 'int'>, <class 'int'>
command_inout DevVoid OK
"""



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
This is an internal PyTango module.
"""

__all__ = ("auto_monitor_init",)

__docformat__ = "restructuredtext"

from tango import AutoTangoMonitor, AutoTangoAllowThreads


def __AutoTangoMonitor__enter__(self):
    self._acquire()
    return self


def __AutoTangoMonitor__exit__(self, *args, **kwargs):
    self._release()


def __init_AutoTangoMonitor():
    AutoTangoMonitor.__enter__ = __AutoTangoMonitor__enter__
    AutoTangoMonitor.__exit__ = __AutoTangoMonitor__exit__


def __AutoTangoAllowThreads__enter__(self):
    return self


def __AutoTangoAllowThreads__exit__(self, *args, **kwargs):
    self._acquire()


def __init_AutoTangoAllowThreads():
    AutoTangoAllowThreads.__enter__ = __AutoTangoAllowThreads__enter__
    AutoTangoAllowThreads.__exit__ = __AutoTangoAllowThreads__exit__


def auto_monitor_init(doc=True):
    __init_AutoTangoMonitor()
    __init_AutoTangoAllowThreads()



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
This is an internal PyTango module.
"""

__all__ = ("Group", "group_init")

__docformat__ = "restructuredtext"

import collections.abc

from tango._tango import __Group as _RealGroup
from tango.utils import _get_command_inout_param
from tango.utils import seq_2_StdStringVector, is_non_str_seq, is_pure_str
from tango.utils import _get_device_fqtrl_if_necessary
from tango.utils import _trace_client
from tango.device_proxy import __init_device_proxy_internals as init_device_proxy


def _apply_to(fn, key):
    if isinstance(key, slice):
        if key.step:
            return [fn(x) for x in range(key.start, key.stop, key.step)]
        else:
            return [fn(x) for x in range(key.start, key.stop)]
    else:
        return fn(key)


def _get_one_item(group, key):
    x = group.get_group(key)
    if x is not None:
        return x
    return group.get_device(key)


# I define Group as a proxy to __Group, where group is the actual
# C++ Tango::Group object. Most functions just call the __group object
# and are defined dynamically in __init_proxy_Group, also copying it's
# documentation strings.
# The proxy is useful for add(group). In this case the parameter 'group'
# becomes useless. With the proxy we make that parameter come to live
# again before returning.
# Another function that needs to be adapted to this is get_group because
# we want to return a Group, not a __Group!
# The get_device method also needs to be adapted in order to properly
# initialize the returned proxy with its python attributes.
class Group:
    """A Tango Group represents a hierarchy of tango devices. The hierarchy
    may have more than one level. The main goal is to group devices with
    same attribute(s)/command(s) to be able to do parallel requests."""

    def __init__(self, name):
        if is_pure_str(name):
            name = _RealGroup(name)
        if not isinstance(name, _RealGroup):
            raise TypeError("Constructor expected receives a str")
        self.__group = name

    def add(self, pattern_subgroup, timeout_ms=-1):
        """
        add(self, device, timeout_ms=-1) -> None
        add(self, device_list, timeout_ms=-1) -> None
        add(self, subgroup, timeout_ms=-1) -> None


        Throws     : TypeError, ArgumentError

            Attaches any device which name matches one of the specified patterns.

            This method first asks the Tango database the list of device names
            matching one the patterns. Devices are then attached to the group in
            the order in which they are returned by the database.

            Any device already present in the hierarchy (i.e. a device belonging to
            the group or to one of its subgroups), is silently ignored but its
            client side timeout is set to timeout_ms milliseconds if timeout_ms
            is different from -1.

            Parameters :
                - device        : (str)a simple device name or a device name pattern (e.g. domain_*/ family/member_*),
                - device_list   : (sequence<str>)  a sequence of these of  a simple device names or
                                    a device name patterns (e.g. domain_*/ family/member_*),
                - subgroup      : (Group) a Group to be attached as subgroup.
                - timeout_ms    : (int) If timeout_ms is different from -1, the client
                                side timeouts of all devices matching the
                                specified patterns are set to timeout_ms
                                milliseconds.
            Return     : None

            Throws     : TypeError, ArgumentError

        """
        if isinstance(pattern_subgroup, Group):
            name = pattern_subgroup.__group.get_name()
            self._add(pattern_subgroup.__group, timeout_ms)
            pattern_subgroup.__group = self.get_group(name)
        else:
            self._add(pattern_subgroup, timeout_ms)

    def _add(self, patterns_or_group, timeout_ms=-1):
        if isinstance(patterns_or_group, _RealGroup):
            items = patterns_or_group
        else:
            if is_pure_str(patterns_or_group):
                items = [patterns_or_group]
            elif is_non_str_seq(patterns_or_group):
                items = patterns_or_group
            else:
                raise TypeError(
                    "Parameter patterns_or_group: Should be Group, "
                    "str or a sequence of strings."
                )

            # If TestContext active, each short TRL is replaced with a fully-qualified
            # TRL, using test server's connection details.  Otherwise, left as-is.
            items = [_get_device_fqtrl_if_necessary(item) for item in items]
            items = seq_2_StdStringVector(items)

        resp = self.__group._add(items, timeout_ms)
        return resp

    def remove(self, patterns, forward=True):
        """
        remove(self, patterns, forward=True) -> None

            Removes any group or device which name matches the specified pattern.

            The pattern parameter can be a group name, a device name or a device
            name pattern (e.g domain_*/family/member_*).

            Since we can have groups with the same name in the hierarchy, a group
            name can be fully qualified to specify which group should be removed.
            Considering the following group:

                ::

                    -> gauges
                    | -> cell-01
                    |     |-> penning
                    |     |    |-> ...
                    |     |-> pirani
                    |          |-> ...
                    | -> cell-02
                    |     |-> penning
                    |     |    |-> ...
                    |     |-> pirani
                    |          |-> ...
                    | -> cell-03
                    |     |-> ...
                    |
                    | -> ...

            A call to gauges->remove("penning") will remove any group named
            "penning" in the hierarchy while gauges->remove("gauges.cell-02.penning")
            will only remove the specified group.

            Parameters :
                - patterns   : (str | sequence<str>) A string with the pattern or a
                               list of patterns.
                - forward    : (bool) If fwd is set to true (the default), the remove
                               request is also forwarded to subgroups. Otherwise,
                               it is only applied to the local set of elements.
                               For instance, the following code remove any
                               stepper motor in the hierarchy:

                                   root_group->remove("*/stepper_motor/*");

            Return     : None

            Throws     :
        """
        if isinstance(patterns, str):
            return self.__group._remove(patterns, forward)
        elif isinstance(patterns, collections.abc.Sequence):
            std_patterns = seq_2_StdStringVector(patterns)
            return self.__group._remove(std_patterns, forward)
        else:
            raise TypeError("Parameter patterns: Should be a str or a sequence of str.")

    def get_device(self, name_or_index):
        proxy = self.__group.get_device(name_or_index)
        if proxy is None:
            raise KeyError(f"Group does not have device {name_or_index}")
        init_device_proxy(proxy)
        return proxy

    def get_group(self, group_name):
        internal = self.__group.get_group(group_name)
        if internal is None:
            return None
        return Group(internal)

    def __contains__(self, pattern):
        return self.contains(pattern)

    def __getitem__(self, key):
        fn = lambda x: _get_one_item(self, x)
        return _apply_to(fn, key)

    def __delitem__(self, key):
        fn = lambda x: self.remove(x)
        return _apply_to(fn, key)

    def __len__(self):
        return self.get_size()

    def __repr__(self):
        return "Group(%s)" % self.get_name()

    @_trace_client
    def command_inout(self, cmd_name, param=None, forward=True):
        """
        command_inout(self, cmd_name, forward=True) -> sequence<GroupCmdReply>
        command_inout(self, cmd_name, param, forward=True) -> sequence<GroupCmdReply>
        command_inout(self, cmd_name, param_list, forward=True) -> sequence<GroupCmdReply>

        Just a shortcut to do:
            self.command_inout_reply(self.command_inout_asynch(...))

        Parameters:
            - cmd_name   : (str) Command name
            - param      : (any) parameter value
            - param_list : (tango.DeviceDataList) sequence of parameters.
                           When given, it's length must match the group size.
            - forward    : (bool) If it is set to true (the default) request is
                            forwarded to subgroups. Otherwise, it is only applied
                            to the local set of devices.

        Return : (sequence<GroupCmdReply>)

        """
        idx = self.command_inout_asynch(cmd_name, param, forward)
        return self.command_inout_reply(idx)

    @_trace_client
    def command_inout_asynch(self, cmd_name, param=None, forward=True):

        if param is None:
            idx = self.__group.command_inout_asynch(
                cmd_name, forget=False, forward=forward
            )
        else:
            arg_in = _get_command_inout_param(self.__group, cmd_name, param)
            idx = self.__group.command_inout_asynch(
                cmd_name, arg_in, forget=False, forward=forward
            )

        return idx

    @_trace_client
    def read_attribute(self, attr_name, forward=True):
        """
        read_attribute(self, attr_name, forward=True) -> sequence<GroupAttrReply>

            Just a shortcut to do:
                self.read_attribute_reply(self.read_attribute_asynch(...))

        """
        idx = self.__group.read_attribute_asynch(attr_name, forward)
        return self.__group.read_attribute_reply(idx)

    @_trace_client
    def read_attributes(self, attr_names, forward=True):
        """
        read_attributes(self, attr_names, forward=True) -> sequence<GroupAttrReply>

            Just a shortcut to do:
                self.read_attributes_reply(self.read_attributes_asynch(...))
        """
        idx = self.__group.read_attributes_asynch(attr_names, forward)
        return self.__group.read_attributes_reply(idx)

    @_trace_client
    def write_attribute(self, attr_name, value, forward=True, multi=False):
        """
        write_attribute(self, attr_name, value, forward=True, multi=False) -> sequence<GroupReply>

            Just a shortcut to do:
                self.write_attribute_reply(self.write_attribute_asynch(...))
        """
        idx = self.__group.write_attribute_asynch(
            attr_name, value, forward=forward, multi=multi
        )
        return self.__group.write_attribute_reply(idx)


def group_init():
    proxy_methods = [
        # 'add',  # Needs to be adapted
        # "command_inout_asynch",  # Needs to be adapted
        "command_inout_reply",
        "contains",
        "disable",
        "enable",
        # 'get_device',  # Needs to be adapted
        "get_device_list",
        "get_fully_qualified_name",
        # 'get_group',   # Needs to be adapted
        "get_name",
        "get_size",
        "is_enabled",
        "name_equals",
        "name_matches",
        "ping",
        "read_attribute_asynch",
        "read_attribute_reply",
        "read_attributes_asynch",
        "read_attributes_reply",
        "remove_all",
        "set_timeout_millis",
        "write_attribute_asynch",
        "write_attribute_reply",
    ]

    def proxy_call_define(fname):
        def fn(self, *args, **kwds):
            return getattr(self._Group__group, fname)(*args, **kwds)

        fn.__doc__ = getattr(_RealGroup, fname).__doc__
        fn.__qualname__ = f"Group.{fname}"
        setattr(Group, fname, _trace_client(fn))

    for fname in proxy_methods:
        proxy_call_define(fname)

        # Group.add.__func__.__doc__ = _RealGroup.add.__doc__
        # Group.get_group.__func__.__doc__ = _RealGroup.get_group.__doc__
        # Group.__doc__ = _RealGroup.__doc__



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
This is an internal PyTango module.
"""

__all__ = ("db_init",)

__docformat__ = "restructuredtext"

import collections.abc

from tango import (
    StdStringVector,
    Database,
    DbDatum,
    DbData,
    DbDevInfo,
    DbDevInfos,
    DbDevExportInfo,
    DbDevExportInfos,
)

from tango.utils import (
    _trace_client,
    is_non_str_seq,
    seq_2_DbDevInfos,
    seq_2_DbDevExportInfos,
    seq_2_DbData,
    DbData_2_dict,
    parameter_2_dbdata,
)

# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# DbDatum extension
# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-


def __DbDatum___setitem(self, k, v):
    self.value_string[k] = v


def __DbDatum___delitem(self, k):
    self.value_string.__delitem__(k)


def __DbDatum_append(self, v):
    self.value_string.append(v)


def __DbDatum_extend(self, v):
    self.value_string.extend(v)


def __DbDatum___imul(self, n):
    self.value_string *= n


def __init_DbDatum():
    DbDatum.__len__ = lambda self: len(self.value_string)
    DbDatum.__getitem__ = lambda self, k: self.value_string[k]
    DbDatum.__setitem__ = __DbDatum___setitem
    DbDatum.__delitem__ = __DbDatum___delitem
    DbDatum.__iter__ = lambda self: self.value_string.__iter__()
    DbDatum.__contains__ = lambda self, v: self.value_string.__contains__(v)
    DbDatum.__add__ = lambda self, seq: self.value_string + seq
    DbDatum.__mul__ = lambda self, n: self.value_string * n
    DbDatum.__imul__ = __DbDatum___imul
    DbDatum.append = __DbDatum_append
    DbDatum.extend = __DbDatum_extend


#    DbDatum.__str__      = __DbDatum___str__
#    DbDatum.__repr__      = __DbDatum___repr__

# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# Database extension
# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-


def __Database__add_server(self, servname, dev_info, with_dserver=False):
    """
    add_server( self, servname, dev_info, with_dserver=False) -> None

            Add a (group of) devices to the database. This is considered as a
            low level call because it may render the database inconsistent
            if it is not used properly.

            If *with_dserver* parameter is set to False (default), this
            call will only register the given dev_info(s). You should include
            in the list of dev_info an entry to the usually hidden **DServer**
            device.

            If *with_dserver* parameter is set to True, the call will add an
            additional **DServer** device if it is not included in the
            *dev_info* parameter.

        Example using *with_dserver=True*::

            dev_info1 = DbDevInfo()
            dev_info1.name = 'my/own/device'
            dev_info1._class = 'MyDevice'
            dev_info1.server = 'MyServer/test'
            db.add_server(dev_info1.server, dev_info1, with_dserver=True)

        Same example using *with_dserver=False*::

            dev_info1 = DbDevInfo()
            dev_info1.name = 'my/own/device'
            dev_info1._class = 'MyDevice'
            dev_info1.server = 'MyServer/test'

            dev_info2 = DbDevInfo()
            dev_info2.name = 'dserver/' + dev_info1.server
            dev_info2._class = 'DServer
            dev_info2.server = dev_info1.server

            dev_info = dev_info1, dev_info2
            db.add_server(dev_info1.server, dev_info)

        .. versionadded:: 8.1.7
            added *with_dserver* parameter

        Parameters :
            - servname : (str) server name
            - dev_info : (sequence<DbDevInfo> | DbDevInfos | DbDevInfo) containing the server device(s) information
            - with_dserver: (bool) whether or not to auto create **DServer** device in server
        Return     : None

        Throws     : ConnectionFailed, CommunicationFailed, DevFailed from device (DB_SQLError)
    """

    if not isinstance(dev_info, collections.abc.Sequence) and not isinstance(
        dev_info, DbDevInfo
    ):
        raise TypeError("Value must be a DbDevInfos, a seq<DbDevInfo> or a DbDevInfo")

    if isinstance(dev_info, DbDevInfos):
        pass
    elif isinstance(dev_info, DbDevInfo):
        dev_info = seq_2_DbDevInfos((dev_info,))
    else:
        dev_info = seq_2_DbDevInfos(dev_info)
    if with_dserver:
        has_dserver = False
        for i in dev_info:
            if i._class == "DServer":
                has_dserver = True
                break
        if not has_dserver:
            dserver_info = DbDevInfo()
            dserver_info.name = "dserver/" + dev_info[0].server
            dserver_info._class = "DServer"
            dserver_info.server = dev_info[0].server
            dev_info.append(dserver_info)
    self._add_server(servname, dev_info)


def __Database__export_server(self, dev_info):
    """
    export_server(self, dev_info) -> None

            Export a group of devices to the database.

        Parameters :
            - devinfo : (sequence<DbDevExportInfo> | DbDevExportInfos | DbDevExportInfo)
                        containing the device(s) to export information
        Return     : None

        Throws     : ConnectionFailed, CommunicationFailed, DevFailed from device (DB_SQLError)
    """

    if not isinstance(dev_info, collections.abc.Sequence) and not isinstance(
        dev_info, DbDevExportInfo
    ):
        raise TypeError(
            "Value must be a DbDevExportInfos, a seq<DbDevExportInfo> or "
            "a DbDevExportInfo"
        )

    if isinstance(dev_info, DbDevExportInfos):
        pass
    elif isinstance(dev_info, DbDevExportInfo):
        dev_info = seq_2_DbDevExportInfos(
            (dev_info),
        )
    else:
        dev_info = seq_2_DbDevExportInfos(dev_info)
    self._export_server(dev_info)


def __generic_get_property(obj_name, value, f):
    new_value = parameter_2_dbdata(value, "value")
    f(obj_name, new_value)
    return new_value, value if isinstance(value, collections.abc.Mapping) else {}


def __Database__generic_get_property(self, obj_name, value, f):
    """internal usage"""

    new_value, ret = __generic_get_property(obj_name, value, f)
    return DbData_2_dict(new_value, ret)


def __Database__generic_put_property(self, obj_name, value, f):
    """internal usage"""
    value = parameter_2_dbdata(value, "value")
    return f(obj_name, value)


def __Database__generic_delete_property(self, obj_name, value, f):
    """internal usage"""
    value = parameter_2_dbdata(value, "value")
    return f(obj_name, value)


def __Database__generic_get_attr_pipe_property(self, obj_name, value, f):
    """internal usage for class or device attribute and pipe properties."""

    new_value, ret = __generic_get_property(obj_name, value, f)
    nb_items = len(new_value)
    i = 0
    while i < nb_items:
        db_datum = new_value[i]
        curr_dict = {}
        ret[db_datum.name] = curr_dict
        nb_props = int(db_datum[0])
        i += 1
        for k in range(nb_props):
            db_datum = new_value[i]
            curr_dict[db_datum.name] = db_datum.value_string
            i += 1

    return ret


def __Database__generic_put_attr_pipe_property(self, obj_name, value, f):
    """internal usage for class or device attribute and pipe properties."""
    new_value = parameter_2_dbdata(value, "value")
    return f(obj_name, new_value)


def __Database__generic_delete_attr_pipe_property(self, obj_name, value, f):
    """internal usage for class or device attribute and pipe properties."""
    if isinstance(value, DbData):
        f(obj_name, value)
    elif is_non_str_seq(value):
        f(obj_name, seq_2_DbData(value))
    elif isinstance(value, collections.abc.Mapping):
        for attr_pipe_name, properties in value.items():
            new_value = DbData()
            new_value.append(DbDatum(attr_pipe_name))
            for prop in properties:
                new_value.append(DbDatum(prop))
            f(obj_name, new_value)
    else:
        raise TypeError(
            "Value must be a string, tango.DbDatum, "
            "tango.DbData, a sequence or a dictionary"
        )


def __Database__put_property(self, obj_name, value):
    """
    put_property(self, obj_name, value) -> None

        Insert or update a list of properties for the specified object.

    Parameters :
        - obj_name : (str) object name
        - value : can be one of the following:

            1. DbDatum - single property data to be inserted
            2. DbData - several property data to be inserted
            3. sequence<DbDatum> - several property data to be inserted
            4. dict<str, DbDatum> - keys are property names and value has data to be inserted
            5. dict<str, obj> - keys are property names and str(obj) is property value
            6. dict<str, seq<str>> - keys are property names and value has data to be inserted
    Return     : None

    Throws     : ConnectionFailed, CommunicationFailed, DevFailed from device (DB_SQLError)
    """

    return __Database__generic_put_property(self, obj_name, value, self._put_property)


def __Database__get_property(self, obj_name, value):
    """
    get_property(self, obj_name, value) -> dict<str, seq<str>>

            Query the database for a list of object (i.e non-device) properties.

        Parameters :
            - obj_name : (str) object name
            - value : can be one of the following:

                1. str [in] - single property data to be fetched
                2. DbDatum [in] - single property data to be fetched
                3. DbData [in,out] - several property data to be fetched
                   In this case (direct C++ API) the DbData will be filled with
                   the property values
                4. sequence<str> [in] - several property data to be fetched
                5. sequence<DbDatum> [in] - several property data to be fetched
                6. dict<str, obj> [in,out] - keys are property names
                   In this case the given dict values will be changed to contain
                   the several property values

        Return     : a dictionary which keys are the property names the value
                     associated with each key being a a sequence of strings being
                     the property value.

        Throws     : ConnectionFailed, CommunicationFailed, DevFailed from device (DB_SQLError)
    """
    return __Database__generic_get_property(self, obj_name, value, self._get_property)


def __Database__get_property_forced(self, obj_name, value):
    return __Database__generic_get_property(
        self, obj_name, value, self._get_property_forced
    )


__Database__get_property_forced.__doc__ = __Database__get_property.__doc__


def __Database__delete_property(self, obj_name, value):
    """
    delete_property(self, obj_name, value) -> None

            Delete a the given of properties for the specified object.

        Parameters :
            - obj_name : (str) object name
            - value : can be one of the following:

                1. str [in] - single property data to be deleted
                2. DbDatum [in] - single property data to be deleted
                3. DbData [in] - several property data to be deleted
                4. sequence<string> [in]- several property data to be deleted
                5. sequence<DbDatum> [in] - several property data to be deleted
                6. dict<str, obj> [in] - keys are property names to be deleted
                   (values are ignored)
                7. dict<str, DbDatum> [in] - several DbDatum.name are property names
                   to be deleted (keys are ignored)
        Return     : None

        Throws     : ConnectionFailed, CommunicationFailed, DevFailed from device (DB_SQLError)
    """
    return __Database__generic_delete_property(
        self, obj_name, value, self._delete_property
    )


def __Database__get_device_property(self, dev_name, value):
    """
    get_device_property(self, dev_name, value) -> dict<str, seq<str>>

        Query the database for a list of device properties.

        Parameters :
            - dev_name : (str) object name
            - value : can be one of the following:

                1. str [in] - single property data to be fetched
                2. DbDatum [in] - single property data to be fetched
                3. DbData [in,out] - several property data to be fetched
                   In this case (direct C++ API) the DbData will be filled with
                   the property values
                4. sequence<str> [in] - several property data to be fetched
                5. sequence<DbDatum> [in] - several property data to be fetched
                6. dict<str, obj> [in,out] - keys are property names
                   In this case the given dict values will be changed to contain
                   the several property values

        Return     : a dictionary which keys are the property names the value
                    associated with each key being a a sequence of strings being the
                    property value.

        Throws     : ConnectionFailed, CommunicationFailed, DevFailed from device (DB_SQLError)
    """
    return __Database__generic_get_property(
        self, dev_name, value, self._get_device_property
    )


def __Database__put_device_property(self, dev_name, value):
    """
    put_device_property(self, dev_name, value) -> None

        Insert or update a list of properties for the specified device.

        Parameters :
            - dev_name : (str) object name
            - value : can be one of the following:

                1. DbDatum - single property data to be inserted
                2. DbData - several property data to be inserted
                3. sequence<DbDatum> - several property data to be inserted
                4. dict<str, DbDatum> - keys are property names and value has data to be inserted
                5. dict<str, obj> - keys are property names and str(obj) is property value
                6. dict<str, seq<str>> - keys are property names and value has data to be inserted
        Return     : None

        Throws     : ConnectionFailed, CommunicationFailed, DevFailed from device (DB_SQLError)
    """
    return __Database__generic_put_property(
        self, dev_name, value, self._put_device_property
    )


def __Database__delete_device_property(self, dev_name, value):
    """
    delete_device_property(self, dev_name, value) -> None

        Delete a the given of properties for the specified device.

        Parameters :
            - dev_name : (str) object name
            - value : can be one of the following:

                1. str [in] - single property data to be deleted
                2. DbDatum [in] - single property data to be deleted
                3. DbData [in] - several property data to be deleted
                4. sequence<str> [in]- several property data to be deleted
                5. sequence<DbDatum> [in] - several property data to be deleted
                6. dict<str, obj> [in] - keys are property names to be deleted (values are ignored)
                7. dict<str, DbDatum> [in] - several DbDatum.name are property names to be deleted (keys are ignored)

        Return     : None

        Throws     : ConnectionFailed, CommunicationFailed, DevFailed from device (DB_SQLError)
    """
    return __Database__generic_delete_property(
        self, dev_name, value, self._delete_device_property
    )


def __Database__get_device_property_list(self, dev_name, wildcard, array=None):
    """
    get_device_property_list(self, dev_name, wildcard, array=None) -> DbData

            Query the database for a list of properties defined for the
            specified device and which match the specified wildcard.
            If array parameter is given, it must be an object implementing de 'append'
            method. If given, it is filled with the matching property names. If not given
            the method returns a new DbDatum containing the matching property names.

        New in PyTango 7.0.0

        Parameters :
            - dev_name : (str) device name
            - wildcard : (str) property name wildcard
            - array : [out] (sequence) (optional) array that
                          will contain the matching property names.
        Return     : if container is None, return is a new DbDatum containing the
                     matching property names. Otherwise returns the given array
                     filled with the property names

        Throws     : ConnectionFailed, CommunicationFailed, DevFailed from device"""
    if array is None:
        return self._get_device_property_list(dev_name, wildcard)
    elif isinstance(array, StdStringVector):
        return self._get_device_property_list(dev_name, wildcard, array)
    elif is_non_str_seq(array):
        res = self._get_device_property_list(dev_name, wildcard)
        for e in res:
            array.append(e)
        return array


def __Database__get_device_attribute_property(self, dev_name, value):
    """
    get_device_attribute_property(self, dev_name, value) -> dict<str, dict<str, seq<str>>>

            Query the database for a list of device attribute properties for the
            specified device. The method returns all the properties for the specified
            attributes.

        Parameters :
            - dev_name : (string) device name
            - value : can be one of the following:

                1. str [in] - single attribute properties to be fetched
                2. DbDatum [in] - single attribute properties to be fetched
                3. DbData [in,out] - several attribute properties to be fetched
                   In this case (direct C++ API) the DbData will be filled with
                   the property values
                4. sequence<str> [in] - several attribute properties to be fetched
                5. sequence<DbDatum> [in] - several attribute properties to be fetched
                6. dict<str, obj> [in,out] - keys are attribute names
                   In this case the given dict values will be changed to contain the
                   several attribute property values

        Return     :  a dictionary which keys are the attribute names the
                             value associated with each key being a another
                             dictionary where keys are property names and value is
                             a DbDatum containing the property value.

        Throws     : ConnectionFailed, CommunicationFailed, DevFailed from device (DB_SQLError)
    """
    return __Database__generic_get_attr_pipe_property(
        self, dev_name, value, self._get_device_attribute_property
    )


def __Database__get_device_pipe_property(self, dev_name, value):
    """
    get_device_pipe_property(self, dev_name, value) -> dict<str, dict<str, seq<str>>>

            Query the database for a list of device pipe properties for the
            specified device. The method returns all the properties for the specified
            pipes.

        Parameters :
            - dev_name : (string) device name
            - value : can be one of the following:

                1. str [in] - single pipe properties to be fetched
                2. DbDatum [in] - single pipe properties to be fetched
                3. DbData [in,out] - several pipe properties to be fetched
                   In this case (direct C++ API) the DbData will be filled with
                   the property values
                4. sequence<str> [in] - several pipe properties to be fetched
                5. sequence<DbDatum> [in] - several pipe properties to be fetched
                6. dict<str, obj> [in,out] - keys are pipe names
                   In this case the given dict values will be changed to contain the
                   several pipe property values

        Return     :  a dictionary which keys are the pipe names the
                             value associated with each key being a another
                             dictionary where keys are property names and value is
                             a DbDatum containing the property value.

        Throws     : ConnectionFailed, CommunicationFailed, DevFailed from device (DB_SQLError)
    """
    return __Database__generic_get_attr_pipe_property(
        self, dev_name, value, self._get_device_pipe_property
    )


def __Database__put_device_attribute_property(self, dev_name, value):
    """
    put_device_attribute_property( self, dev_name, value) -> None

            Insert or update a list of properties for the specified device.

        Parameters :
            - dev_name : (str) device name
            - value : can be one of the following:

                1. DbData - several property data to be inserted
                2. sequence<DbDatum> - several property data to be inserted
                3. dict<str, dict<str, obj>> keys are attribute names and value being another
                   dictionary which keys are the attribute property names and the value
                   associated with each key being seq<str> or tango.DbDatum.

        Return     : None

        Throws     : ConnectionFailed, CommunicationFailed, DevFailed from device (DB_SQLError)
    """
    return __Database__generic_put_attr_pipe_property(
        self, dev_name, value, self._put_device_attribute_property
    )


def __Database__put_device_pipe_property(self, dev_name, value):
    """
    put_device_pipe_property( self, dev_name, value) -> None

            Insert or update a list of properties for the specified device.

        Parameters :
            - dev_name : (str) device name
            - value : can be one of the following:

                1. DbData - several property data to be inserted
                2. sequence<DbDatum> - several property data to be inserted
                3. dict<str, dict<str, obj>> keys are pipe names and value being another
                   dictionary which keys are the pipe property names and the value
                   associated with each key being seq<str> or tango.DbDatum.

        Return     : None

        Throws     : ConnectionFailed, CommunicationFailed, DevFailed from device (DB_SQLError)
    """
    return __Database__generic_put_attr_pipe_property(
        self, dev_name, value, self._put_device_pipe_property
    )


def __Database__delete_device_attribute_property(self, dev_name, value):
    """
    delete_device_attribute_property(self, dev_name, value) -> None

            Delete a list of attribute properties for the specified device.

        Parameters :
            - devname : (str) device name
            - propnames : can be one of the following:

                1. DbData [in] - several property data to be deleted
                2. sequence<str> [in]- several property data to be deleted
                3. sequence<DbDatum> [in] - several property data to be deleted
                4. dict<str, seq<str>> with each key an attribute name and the value a list of attribute property names to delete from that attribute

        Return     : None

        Throws     : ConnectionFailed, CommunicationFailed, DevFailed from device (DB_SQLError)
    """
    return __Database__generic_delete_attr_pipe_property(
        self, dev_name, value, self._delete_device_attribute_property
    )


def __Database__delete_device_pipe_property(self, dev_name, value):
    """
    delete_device_pipe_property(self, dev_name, value) -> None

            Delete a list of pipe properties for the specified device.

        Parameters :
            - devname : (string) device name
            - propnames : can be one of the following:

                1. DbData [in] - several property data to be deleted
                2. sequence<str> [in]- several property data to be deleted
                3. sequence<DbDatum> [in] - several property data to be deleted
                3. dict<str, seq<str>> with each key a pipe name and the value a list of pipe property names to delete from that pipe

        Return     : None

        Throws     : ConnectionFailed, CommunicationFailed, DevFailed from device (DB_SQLError)
    """
    return __Database__generic_delete_attr_pipe_property(
        self, dev_name, value, self._delete_device_pipe_property
    )


def __Database__get_class_property(self, class_name, value):
    """
    get_class_property(self, class_name, value) -> dict<str, seq<str>>

            Query the database for a list of class properties.

        Parameters :
            - class_name : (str) class name
            - value : can be one of the following:

                1. str [in] - single property data to be fetched
                2. tango.DbDatum [in] - single property data to be fetched
                3. tango.DbData [in,out] - several property data to be fetched
                   In this case (direct C++ API) the DbData will be filled with
                   the property values
                4. sequence<str> [in] - several property data to be fetched
                5. sequence<DbDatum> [in] - several property data to be fetched
                6. dict<str, obj> [in,out] - keys are property names
                   In this case the given dict values will be changed to contain
                   the several property values

        Return     : a dictionary which keys are the property names the value
                     associated with each key being a a sequence of strings being the
                     property value.

        Throws     : ConnectionFailed, CommunicationFailed, DevFailed from device (DB_SQLError)
    """
    return __Database__generic_get_property(
        self, class_name, value, self._get_class_property
    )


def __Database__put_class_property(self, class_name, value):
    """
    put_class_property(self, class_name, value) -> None

            Insert or update a list of properties for the specified class.

        Parameters :
            - class_name : (str) class name
            - value : can be one of the following:

                1. DbDatum - single property data to be inserted
                2. DbData - several property data to be inserted
                3. sequence<DbDatum> - several property data to be inserted
                4. dict<str, DbDatum> - keys are property names and value has data to be inserted
                5. dict<str, obj> - keys are property names and str(obj) is property value
                6. dict<str, seq<str>> - keys are property names and value has data to be inserted
        Return     : None

        Throws     : ConnectionFailed, CommunicationFailed, DevFailed from device (DB_SQLError)
    """
    return __Database__generic_put_property(
        self, class_name, value, self._put_class_property
    )


def __Database__delete_class_property(self, class_name, value):
    """
    delete_class_property(self, class_name, value) -> None

            Delete a the given of properties for the specified class.

        Parameters :
            - class_name : (str) class name
            - value : can be one of the following:

                1. str [in] - single property data to be deleted
                2. DbDatum [in] - single property data to be deleted
                3. DbData [in] - several property data to be deleted
                4. sequence<str> [in]- several property data to be deleted
                5. sequence<DbDatum> [in] - several property data to be deleted
                6. dict<str, obj> [in] - keys are property names to be deleted
                   (values are ignored)
                7. dict<str, DbDatum> [in] - several DbDatum.name are property names
                   to be deleted (keys are ignored)

        Return     : None

        Throws     : ConnectionFailed, CommunicationFailed, DevFailed from device (DB_SQLError)
    """
    return __Database__generic_delete_property(
        self, class_name, value, self._delete_class_property
    )


def __Database__get_class_attribute_property(self, class_name, value):
    """
    get_class_attribute_property( self, class_name, value) -> dict<str, dict<str, seq<str>>

            Query the database for a list of class attribute properties for the
            specified class. The method returns all the properties for the specified
            attributes.

        Parameters :
            - class_name : (str) class name
            - propnames : can be one of the following:

                1. str [in] - single attribute properties to be fetched
                2. DbDatum [in] - single attribute properties to be fetched
                3. DbData [in,out] - several attribute properties to be fetched
                   In this case (direct C++ API) the DbData will be filled with the property
                   values
                4. sequence<str> [in] - several attribute properties to be fetched
                5. sequence<DbDatum> [in] - several attribute properties to be fetched
                6. dict<str, obj> [in,out] - keys are attribute names
                   In this case the given dict values will be changed to contain the several
                   attribute property values

        Return     : a dictionary which keys are the attribute names the
                     value associated with each key being a another
                     dictionary where keys are property names and value is
                     a sequence of strings being the property value.

        Throws     : ConnectionFailed, CommunicationFailed, DevFailed from device (DB_SQLError)
    """
    return __Database__generic_get_attr_pipe_property(
        self, class_name, value, self._get_class_attribute_property
    )


def __Database__put_class_attribute_property(self, class_name, value):
    """
    put_class_attribute_property(self, class_name, value) -> None

            Insert or update a list of properties for the specified class.

        Parameters :
            - class_name : (str) class name
            - propdata : can be one of the following:

                1. tango.DbData - several property data to be inserted
                2. sequence<DbDatum> - several property data to be inserted
                3. dict<str, dict<str, obj>> keys are attribute names and value
                   being another dictionary which keys are the attribute property
                   names and the value associated with each key being seq<str> or
                   tango.DbDatum

        Return     : None

        Throws     : ConnectionFailed, CommunicationFailed, DevFailed from device (DB_SQLError)
    """
    return __Database__generic_put_attr_pipe_property(
        self, class_name, value, self._put_class_attribute_property
    )


def __Database__delete_class_attribute_property(self, class_name, value):
    """
    delete_class_attribute_property(self, class_name, value) -> None

            Delete a list of attribute properties for the specified class.

        Parameters :
            - class_name : (str) class name
            - propnames : can be one of the following:

                1. DbData [in] - several property data to be deleted
                2. sequence<str> [in]- several property data to be deleted
                3. sequence<DbDatum> [in] - several property data to be deleted
                4. dict<str, seq<str>> keys are attribute names and value being a
                   list of attribute property names

        Return     : None

        Throws     : ConnectionFailed, CommunicationFailed
                     DevFailed from device (DB_SQLError)"""
    return __Database__generic_delete_attr_pipe_property(
        self, class_name, value, self._delete_class_attribute_property
    )


def __Database__get_service_list(self, filter=".*"):
    import re

    data = self.get_property("CtrlSystem", "Services")
    res = {}
    filter_re = re.compile(filter)
    for service in data["Services"]:
        service_name, service_value = service.split(":")
        if filter_re.match(service_name) is not None:
            res[service_name] = service_value
    return res


def __Database__str(self):
    return f"Database({self.get_db_host()}, {self.get_db_port()})"


def __init_Database():
    Database.add_server = _trace_client(__Database__add_server)
    Database.export_server = _trace_client(__Database__export_server)
    Database.put_property = _trace_client(__Database__put_property)
    Database.get_property = _trace_client(__Database__get_property)
    Database.get_property_forced = _trace_client(__Database__get_property_forced)
    Database.delete_property = _trace_client(__Database__delete_property)
    Database.get_device_property = _trace_client(__Database__get_device_property)
    Database.put_device_property = _trace_client(__Database__put_device_property)
    Database.delete_device_property = _trace_client(__Database__delete_device_property)
    Database.get_device_property_list = _trace_client(
        __Database__get_device_property_list
    )
    Database.get_device_attribute_property = _trace_client(
        __Database__get_device_attribute_property
    )
    Database.put_device_attribute_property = _trace_client(
        __Database__put_device_attribute_property
    )
    Database.delete_device_attribute_property = _trace_client(
        __Database__delete_device_attribute_property
    )
    Database.get_class_property = _trace_client(__Database__get_class_property)
    Database.put_class_property = _trace_client(__Database__put_class_property)
    Database.delete_class_property = _trace_client(__Database__delete_class_property)
    Database.get_class_attribute_property = _trace_client(
        __Database__get_class_attribute_property
    )
    Database.put_class_attribute_property = _trace_client(
        __Database__put_class_attribute_property
    )
    Database.delete_class_attribute_property = _trace_client(
        __Database__delete_class_attribute_property
    )
    Database.get_service_list = _trace_client(__Database__get_service_list)
    Database.__str__ = __Database__str
    Database.__repr__ = __Database__str


def db_init():
    __init_DbDatum()
    __init_Database()



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
"""
This crashes under PyTango 9.3.3 and cppTango 9.3.4.  Possibly other versions too.
"""

from tango.test_context import DeviceTestContext
import tango
import time


class Device1(tango.server.Device):
    pass


with DeviceTestContext(Device1, process=True) as proxy:
    cb = tango.utils.EventCallback()
    eid = proxy.subscribe_event("state", tango.EventType.ATTR_CONF_EVENT, cb)
    sleep = 12  # sleep >~11 ==> SIGSEGV ; sleep<10 ==> exit OK
    for i in range(sleep):
        time.sleep(1)
        print(i + 1, end=", ")
    print("\nbefore unsubscribe")
    proxy.unsubscribe_event(eid)  # <-- SEGFAULT here
    print("after unsubscribe")



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
This is an internal PyTango module.
"""

__all__ = ("encoded_attribute_init",)

__docformat__ = "restructuredtext"

import numpy as np

from tango._tango import EncodedAttribute, ExtractAs, _ImageFormat

from tango.utils import is_pure_str, is_seq

_allowed_extract = (ExtractAs.Numpy, ExtractAs.String, ExtractAs.Tuple, ExtractAs.List)


def __EncodedAttribute_encode_jpeg_gray8(self, gray8, width=0, height=0, quality=100.0):
    """Encode a 8 bit grayscale image as JPEG format

        :param gray8: an object containning image information
        :type gray8: :py:obj:`str` or :class:`numpy.ndarray` or seq< seq<element> >
        :param width: image width. **MUST** be given if gray8 is a string or
                      if it is a :class:`numpy.ndarray` with ndims != 2.
                      Otherwise it is calculated internally.
        :type width: :py:obj:`int`
        :param height: image height. **MUST** be given if gray8 is a string
                       or if it is a :class:`numpy.ndarray` with ndims != 2.
                       Otherwise it is calculated internally.
        :type height: :py:obj:`int`
        :param quality: Quality of JPEG (0=poor quality 100=max quality) (default is 100.0)
        :type quality: :py:obj:`float`

    .. note::
        When :class:`numpy.ndarray` is given:

            - gray8 **MUST** be CONTIGUOUS, ALIGNED
            - if gray8.ndims != 2, width and height **MUST** be given and
              gray8.nbytes **MUST** match width*height
            - if gray8.ndims == 2, gray8.itemsize **MUST** be 1 (typically,
              gray8.dtype is one of `numpy.dtype.byte`, `numpy.dtype.ubyte`,
              `numpy.dtype.int8` or `numpy.dtype.uint8`)

    Example::

        def read_myattr(self):
            enc = tango.EncodedAttribute()
            data = numpy.arange(100, dtype=numpy.byte)
            data = numpy.array((data,data,data))
            enc.encode_jpeg_gray8(data)
            return enc
    """
    self._generic_encode_gray8(
        gray8,
        width=width,
        height=height,
        quality=quality,
        format=_ImageFormat.JpegImage,
    )


def __EncodedAttribute_encode_gray8(self, gray8, width=0, height=0):
    """Encode a 8 bit grayscale image (no compression)

        :param gray8: an object containning image information
        :type gray8: :py:obj:`str` or :class:`numpy.ndarray` or seq< seq<element> >
        :param width: image width. **MUST** be given if gray8 is a string or
                      if it is a :class:`numpy.ndarray` with ndims != 2.
                      Otherwise it is calculated internally.
        :type width: :py:obj:`int`
        :param height: image height. **MUST** be given if gray8 is a string
                       or if it is a :class:`numpy.ndarray` with ndims != 2.
                       Otherwise it is calculated internally.
        :type height: :py:obj:`int`

    .. note::
        When :class:`numpy.ndarray` is given:

            - gray8 **MUST** be CONTIGUOUS, ALIGNED
            - if gray8.ndims != 2, width and height **MUST** be given and
              gray8.nbytes **MUST** match width*height
            - if gray8.ndims == 2, gray8.itemsize **MUST** be 1 (typically,
              gray8.dtype is one of `numpy.dtype.byte`, `numpy.dtype.ubyte`,
              `numpy.dtype.int8` or `numpy.dtype.uint8`)

    Example::

        def read_myattr(self):
            enc = tango.EncodedAttribute()
            data = numpy.arange(100, dtype=numpy.byte)
            data = numpy.array((data,data,data))
            enc.encode_gray8(data)
            return enc
    """
    self._generic_encode_gray8(
        gray8, width=width, height=height, format=_ImageFormat.RawImage
    )


def __EncodedAttribute_generic_encode_gray8(
    self, gray8, width=0, height=0, quality=0, format=_ImageFormat.RawImage
):
    """Internal usage only"""
    if not is_seq(gray8):
        raise TypeError(
            "Expected sequence (str, numpy.ndarray, list, tuple "
            "or bytearray) as first argument"
        )

    is_str = is_pure_str(gray8)
    if is_str:
        if not width or not height:
            raise ValueError(
                "When giving a string as data, you must also supply width and height"
            )

    if isinstance(gray8, np.ndarray):
        if gray8.ndim != 2:
            if not width or not height:
                raise ValueError(
                    "When giving a non 2D numpy array, width and "
                    "height must be supplied"
                )
            if gray8.nbytes != width * height:
                raise ValueError("numpy array size mismatch")
        else:
            if gray8.itemsize != 1:
                raise TypeError("Expected numpy array with itemsize == 1")
        if not gray8.flags.c_contiguous:
            raise TypeError(
                "Currently, only contiguous, aligned numpy arrays are supported"
            )
        if not gray8.flags.aligned:
            raise TypeError(
                "Currently, only contiguous, aligned numpy arrays are supported"
            )

    if not is_str and (not width or not height):
        height = len(gray8)
        if height < 1:
            raise IndexError("Expected sequence with at least one row")

        row0 = gray8[0]
        if not is_seq(row0):
            raise IndexError(
                "Expected sequence (str, numpy.ndarray, list, tuple or "
                "bytearray) inside a sequence"
            )
        width = len(row0)

    if format == _ImageFormat.RawImage:
        self._encode_gray8(gray8, width, height)
    elif format == _ImageFormat.JpegImage:
        self._encode_jpeg_gray8(gray8, width, height, quality)


def __EncodedAttribute_encode_gray16(self, gray16, width=0, height=0):
    """Encode a 16 bit grayscale image (no compression)

        :param gray16: an object containning image information
        :type gray16: :py:obj:`str` or :py:obj:`buffer` or :class:`numpy.ndarray` or seq< seq<element> >
        :param width: image width. **MUST** be given if gray16 is a string or
                      if it is a :class:`numpy.ndarray` with ndims != 2.
                      Otherwise it is calculated internally.
        :type width: :py:obj:`int`
        :param height: image height. **MUST** be given if gray16 is a string
                       or if it is a :class:`numpy.ndarray` with ndims != 2.
                       Otherwise it is calculated internally.
        :type height: :py:obj:`int`

    .. note::
        When :class:`numpy.ndarray` is given:

            - gray16 **MUST** be CONTIGUOUS, ALIGNED
            - if gray16.ndims != 2, width and height **MUST** be given and
              gray16.nbytes/2 **MUST** match width*height
            - if gray16.ndims == 2, gray16.itemsize **MUST** be 2 (typically,
              gray16.dtype is one of `numpy.dtype.int16`, `numpy.dtype.uint16`,
              `numpy.dtype.short` or `numpy.dtype.ushort`)

    Example::

        def read_myattr(self):
            enc = tango.EncodedAttribute()
            data = numpy.arange(100, dtype=numpy.int16)
            data = numpy.array((data,data,data))
            enc.encode_gray16(data)
            return enc
    """
    if not is_seq(gray16):
        raise TypeError(
            "Expected sequence (str, numpy.ndarray, list, tuple "
            "or bytearray) as first argument"
        )

    is_str = is_pure_str(gray16)
    if is_str:
        if not width or not height:
            raise ValueError(
                "When giving a string as data, you must also supply width and height"
            )

    if isinstance(gray16, np.ndarray):
        if gray16.ndim != 2:
            if not width or not height:
                raise ValueError(
                    "When giving a non 2D numpy array, width and "
                    "height must be supplied"
                )
            if gray16.nbytes / 2 != width * height:
                raise ValueError("numpy array size mismatch")
        else:
            if gray16.itemsize != 2:
                raise TypeError("Expected numpy array with itemsize == 2")
        if not gray16.flags.c_contiguous:
            raise TypeError(
                "Currently, only contiguous, aligned numpy arrays are supported"
            )
        if not gray16.flags.aligned:
            raise TypeError(
                "Currently, only contiguous, aligned numpy arrays are supported"
            )

    if not is_str and (not width or not height):
        height = len(gray16)
        if height < 1:
            raise IndexError("Expected sequence with at least one row")

        row0 = gray16[0]
        if not is_seq(row0):
            raise IndexError(
                "Expected sequence (str, numpy.ndarray, list, tuple or "
                "bytearray) inside a sequence"
            )
        width = len(row0)
        if is_pure_str(row0) or type(row0) is bytearray:
            width /= 2

    self._encode_gray16(gray16, width, height)


def __EncodedAttribute_encode_jpeg_rgb24(self, rgb24, width=0, height=0, quality=100.0):
    """Encode a 24 bit rgb color image as JPEG format.

        :param rgb24: an object containning image information
        :type rgb24: :py:obj:`str` or :class:`numpy.ndarray` or seq< seq<element> >
        :param width: image width. **MUST** be given if rgb24 is a string or
                      if it is a :class:`numpy.ndarray` with ndims != 3.
                      Otherwise it is calculated internally.
        :type width: :py:obj:`int`
        :param height: image height. **MUST** be given if rgb24 is a string
                       or if it is a :class:`numpy.ndarray` with ndims != 3.
                       Otherwise it is calculated internally.
        :type height: :py:obj:`int`
        :param quality: Quality of JPEG (0=poor quality 100=max quality) (default is 100.0)
        :type quality: :py:obj:`float`

    .. note::
        When :class:`numpy.ndarray` is given:

            - rgb24 **MUST** be CONTIGUOUS, ALIGNED
            - if rgb24.ndims != 3, width and height **MUST** be given and
              rgb24.nbytes/3 **MUST** match width*height
            - if rgb24.ndims == 3, rgb24.itemsize **MUST** be 1 (typically,
              rgb24.dtype is one of `numpy.dtype.byte`, `numpy.dtype.ubyte`,
              `numpy.dtype.int8` or `numpy.dtype.uint8`) and shape **MUST** be
              (height, width, 3)

    Example::

        def read_myattr(self):
            enc = tango.EncodedAttribute()
            # create an 'image' where each pixel is R=0x01, G=0x01, B=0x01
            arr = numpy.ones((10,10,3), dtype=numpy.uint8)
            enc.encode_jpeg_rgb24(data)
            return enc
    """
    self._generic_encode_rgb24(
        rgb24,
        width=width,
        height=height,
        quality=quality,
        format=_ImageFormat.JpegImage,
    )


def __EncodedAttribute_encode_rgb24(self, rgb24, width=0, height=0):
    """Encode a 24 bit color image (no compression)

        :param rgb24: an object containning image information
        :type rgb24: :py:obj:`str` or :class:`numpy.ndarray` or seq< seq<element> >
        :param width: image width. **MUST** be given if rgb24 is a string or
                      if it is a :class:`numpy.ndarray` with ndims != 3.
                      Otherwise it is calculated internally.
        :type width: :py:obj:`int`
        :param height: image height. **MUST** be given if rgb24 is a string
                       or if it is a :class:`numpy.ndarray` with ndims != 3.
                       Otherwise it is calculated internally.
        :type height: :py:obj:`int`

    .. note::
        When :class:`numpy.ndarray` is given:

            - rgb24 **MUST** be CONTIGUOUS, ALIGNED
            - if rgb24.ndims != 3, width and height **MUST** be given and
              rgb24.nbytes/3 **MUST** match width*height
            - if rgb24.ndims == 3, rgb24.itemsize **MUST** be 1 (typically,
              rgb24.dtype is one of `numpy.dtype.byte`, `numpy.dtype.ubyte`,
              `numpy.dtype.int8` or `numpy.dtype.uint8`) and shape **MUST** be
              (height, width, 3)

    Example::

        def read_myattr(self):
            enc = tango.EncodedAttribute()
            # create an 'image' where each pixel is R=0x01, G=0x01, B=0x01
            arr = numpy.ones((10,10,3), dtype=numpy.uint8)
            enc.encode_rgb24(data)
            return enc
    """
    self._generic_encode_rgb24(
        rgb24, width=width, height=height, format=_ImageFormat.RawImage
    )


def __EncodedAttribute_generic_encode_rgb24(
    self, rgb24, width=0, height=0, quality=0, format=_ImageFormat.RawImage
):
    """Internal usage only"""
    if not is_seq(rgb24):
        raise TypeError(
            "Expected sequence (str, numpy.ndarray, list, tuple "
            "or bytearray) as first argument"
        )

    is_str = is_pure_str(rgb24)
    if is_str:
        if not width or not height:
            raise ValueError(
                "When giving a string as data, you must also supply width and height"
            )

    if isinstance(rgb24, np.ndarray):
        if rgb24.ndim != 3:
            if not width or not height:
                raise ValueError(
                    "When giving a non 2D numpy array, width and "
                    "height must be supplied"
                )
            if rgb24.nbytes / 3 != width * height:
                raise ValueError("numpy array size mismatch")
        else:
            if rgb24.itemsize != 1:
                raise TypeError("Expected numpy array with itemsize == 1")
        if not rgb24.flags.c_contiguous:
            raise TypeError(
                "Currently, only contiguous, aligned numpy arrays are supported"
            )
        if not rgb24.flags.aligned:
            raise TypeError(
                "Currently, only contiguous, aligned numpy arrays are supported"
            )

    if not is_str and (not width or not height):
        height = len(rgb24)
        if height < 1:
            raise IndexError("Expected sequence with at least one row")

        row0 = rgb24[0]
        if not is_seq(row0):
            raise IndexError(
                "Expected sequence (str, numpy.ndarray, list, tuple or "
                "bytearray) inside a sequence"
            )
        width = len(row0)
        if is_pure_str(row0) or type(row0) is bytearray:
            width /= 3
    if format == _ImageFormat.RawImage:
        self._encode_rgb24(rgb24, width, height)
    elif format == _ImageFormat.JpegImage:
        self._encode_jpeg_rgb24(rgb24, width, height, quality)


def __EncodedAttribute_encode_jpeg_rgb32(self, rgb32, width=0, height=0, quality=100.0):
    """Encode a 32 bit rgb color image as JPEG format.

        :param rgb32: an object containning image information
        :type rgb32: :py:obj:`str` or :class:`numpy.ndarray` or seq< seq<element> >
        :param width: image width. **MUST** be given if rgb32 is a string or
                      if it is a :class:`numpy.ndarray` with ndims != 2.
                      Otherwise it is calculated internally.
        :type width: :py:obj:`int`
        :param height: image height. **MUST** be given if rgb32 is a string
                       or if it is a :class:`numpy.ndarray` with ndims != 2.
                       Otherwise it is calculated internally.
        :type height: :py:obj:`int`

    .. note::
        When :class:`numpy.ndarray` is given:

            - rgb32 **MUST** be CONTIGUOUS, ALIGNED
            - if rgb32.ndims != 2, width and height **MUST** be given and
              rgb32.nbytes/4 **MUST** match width*height
            - if rgb32.ndims == 2, rgb32.itemsize **MUST** be 4 (typically,
              rgb32.dtype is one of `numpy.dtype.int32`, `numpy.dtype.uint32`)

    .. note::

        Encoding with transparency information required cppTango built with `TANGO_USE_JPEG` options
        see installation instructions of cppTango

    Example::

        def read_myattr(self):
            enc = tango.EncodedAttribute()
            data = numpy.arange(100, dtype=numpy.int32)
            data = numpy.array((data,data,data))
            enc.encode_jpeg_rgb32(data)
            return enc
    """
    if not is_seq(rgb32):
        raise TypeError(
            "Expected sequence (str, numpy.ndarray, list, tuple "
            "or bytearray) as first argument"
        )

    is_str = is_pure_str(rgb32)
    if is_str:
        if not width or not height:
            raise ValueError(
                "When giving a string as data, you must also supply width and height"
            )

    if isinstance(rgb32, np.ndarray):
        if rgb32.ndim != 2:
            if not width or not height:
                raise ValueError(
                    "When giving a non 2D numpy array, width and "
                    "height must be supplied"
                )
            if rgb32.nbytes / 4 != width * height:
                raise ValueError("numpy array size mismatch")
        else:
            if rgb32.itemsize != 4:
                raise TypeError("Expected numpy array with itemsize == 4")
        if not rgb32.flags.c_contiguous:
            raise TypeError(
                "Currently, only contiguous, aligned numpy arrays are supported"
            )
        if not rgb32.flags.aligned:
            raise TypeError(
                "Currently, only contiguous, aligned numpy arrays are supported"
            )

    if not is_str and (not width or not height):
        height = len(rgb32)
        if height < 1:
            raise IndexError("Expected sequence with at least one row")

        row0 = rgb32[0]
        if not is_seq(row0):
            raise IndexError(
                "Expected sequence (str, numpy.ndarray, list, tuple or "
                "bytearray) inside a sequence"
            )
        width = len(row0)
        if is_pure_str(row0) or type(row0) is bytearray:
            width /= 4

    self._encode_jpeg_rgb32(rgb32, width, height, quality)


def __EncodedAttribute_decode_gray8(self, da, extract_as=ExtractAs.Numpy):
    """Decode a 8 bits grayscale image (JPEG_GRAY8 or GRAY8) and returns a 8 bits gray scale image.

     :param da: :class:`DeviceAttribute` that contains the image
     :type da: :class:`DeviceAttribute`
     :param extract_as: defaults to ExtractAs.Numpy
     :type extract_as: ExtractAs
     :return: the decoded data

     - In case String string is choosen as extract method, a tuple is returned:
         width<int>, height<int>, buffer<str>
     - In case Numpy is choosen as extract method, a :class:`numpy.ndarray` is
       returned with ndim=2, shape=(height, width) and dtype=numpy.uint8.
     - In case Tuple or List are choosen, a tuple<tuple<int>> or list<list<int>>
       is returned.

    .. warning::
        The PyTango calls that return a :class:`DeviceAttribute`
        (like :meth:`DeviceProxy.read_attribute` or :meth:`DeviceProxy.command_inout`)
        automatically extract the contents by default. This method requires
        that the given :class:`DeviceAttribute` is obtained from a
        call which **DOESN'T** extract the contents. Example::

            dev = tango.DeviceProxy("a/b/c")
            da = dev.read_attribute("my_attr", extract_as=tango.ExtractAs.Nothing)
            enc = tango.EncodedAttribute()
            data = enc.decode_gray8(da)
    """
    if hasattr(da, "value"):
        raise TypeError(
            "DeviceAttribute argument must have been obtained from "
            "a call which doesn't extract the contents"
        )
    if extract_as not in _allowed_extract:
        raise TypeError("extract_as must be one of Numpy, String, Tuple, List")
    return self._decode_gray8(da, extract_as)


def __EncodedAttribute_decode_gray16(self, da, extract_as=ExtractAs.Numpy):
    """Decode a 16 bits grayscale image (GRAY16) and returns a 16 bits gray scale image.

     :param da: :class:`DeviceAttribute` that contains the image
     :type da: :class:`DeviceAttribute`
     :param extract_as: defaults to ExtractAs.Numpy
     :type extract_as: ExtractAs
     :return: the decoded data

     - In case String string is choosen as extract method, a tuple is returned:
         width<int>, height<int>, buffer<str>
     - In case Numpy is choosen as extract method, a :class:`numpy.ndarray` is
       returned with ndim=2, shape=(height, width) and dtype=numpy.uint16.
     - In case Tuple or List are choosen, a tuple<tuple<int>> or list<list<int>>
       is returned.

    .. warning::
        The PyTango calls that return a :class:`DeviceAttribute`
        (like :meth:`DeviceProxy.read_attribute` or :meth:`DeviceProxy.command_inout`)
        automatically extract the contents by default. This method requires
        that the given :class:`DeviceAttribute` is obtained from a
        call which **DOESN'T** extract the contents. Example::

            dev = tango.DeviceProxy("a/b/c")
            da = dev.read_attribute("my_attr", extract_as=tango.ExtractAs.Nothing)
            enc = tango.EncodedAttribute()
            data = enc.decode_gray16(da)
    """
    if hasattr(da, "value"):
        raise TypeError(
            "DeviceAttribute argument must have been obtained from "
            "a call which doesn't extract the contents"
        )
    if extract_as not in _allowed_extract:
        raise TypeError("extract_as must be one of Numpy, String, Tuple, List")
    return self._decode_gray16(da, extract_as)


def __EncodedAttribute_decode_rgb32(self, da, extract_as=ExtractAs.Numpy):
    """Decode a color image (JPEG_RGB or RGB24) and returns a 32 bits RGB image.

     :param da: :class:`DeviceAttribute` that contains the image
     :type da: :class:`DeviceAttribute`
     :param extract_as: defaults to ExtractAs.Numpy
     :type extract_as: ExtractAs
     :return: the decoded data

     - In case String string is choosen as extract method, a tuple is returned:
         width<int>, height<int>, buffer<str>
     - In case Numpy is choosen as extract method, a :class:`numpy.ndarray` is
       returned with ndim=2, shape=(height, width) and dtype=numpy.uint32.
     - In case Tuple or List are choosen, a tuple<tuple<int>> or list<list<int>>
       is returned.

    .. warning::
        The PyTango calls that return a :class:`DeviceAttribute`
        (like :meth:`DeviceProxy.read_attribute` or :meth:`DeviceProxy.command_inout`)
        automatically extract the contents by default. This method requires
        that the given :class:`DeviceAttribute` is obtained from a
        call which **DOESN'T** extract the contents. Example::

            dev = tango.DeviceProxy("a/b/c")
            da = dev.read_attribute("my_attr", extract_as=tango.ExtractAs.Nothing)
            enc = tango.EncodedAttribute()
            data = enc.decode_rgb32(da)
    """
    if hasattr(da, "value"):
        raise TypeError(
            "DeviceAttribute argument must have been obtained from "
            "a call which doesn't extract the contents"
        )
    if extract_as not in _allowed_extract:
        raise TypeError("extract_as must be one of Numpy, String, Tuple, List")
    return self._decode_rgb32(da, extract_as)


def encoded_attribute_init():
    EncodedAttribute._generic_encode_gray8 = __EncodedAttribute_generic_encode_gray8
    EncodedAttribute.encode_gray8 = __EncodedAttribute_encode_gray8
    EncodedAttribute.encode_jpeg_gray8 = __EncodedAttribute_encode_jpeg_gray8
    EncodedAttribute.encode_gray16 = __EncodedAttribute_encode_gray16
    EncodedAttribute._generic_encode_rgb24 = __EncodedAttribute_generic_encode_rgb24
    EncodedAttribute.encode_rgb24 = __EncodedAttribute_encode_rgb24
    EncodedAttribute.encode_jpeg_rgb24 = __EncodedAttribute_encode_jpeg_rgb24
    EncodedAttribute.encode_jpeg_rgb32 = __EncodedAttribute_encode_jpeg_rgb32
    EncodedAttribute.decode_gray8 = __EncodedAttribute_decode_gray8
    EncodedAttribute.decode_gray16 = __EncodedAttribute_decode_gray16
    EncodedAttribute.decode_rgb32 = __EncodedAttribute_decode_rgb32



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
"""
This is an internal PyTango module. It completes the binding of
:class:`tango.AttributeProxy`.

To access these members use directly :mod:`tango` module and NOT
tango.attribute_proxy.
"""

from tango._tango import (
    DeviceProxy,
    DevFailed,
    Except,
)
from tango._tango import __AttributeProxy as _AttributeProxy, DbDatum, DbData
from tango.utils import parameter_2_dbdata, get_property_from_db
from tango.utils import _get_device_fqtrl_if_necessary
from tango.green import green, get_green_mode
from tango.utils import _trace_client
from tango.device_proxy import __init_device_proxy_internals as init_device_proxy

__all__ = ("AttributeProxy", "attribute_proxy_init", "get_attribute_proxy")


@green(consume_green_mode=False)
@_trace_client
def get_attribute_proxy(*args, green_mode=None):
    """
    get_attribute_proxy(self, full_attr_name, green_mode=None, wait=True, timeout=True) -> AttributeProxy
    get_attribute_proxy(self, device_proxy, attr_name, green_mode=None, wait=True, timeout=True) -> AttributeProxy

    Returns a new :class:`~tango.AttributeProxy`.
    There is no difference between using this function and the direct
    :class:`~tango.AttributeProxy` constructor if you use the default kwargs.

    The added value of this function becomes evident when you choose a green_mode
    to be *Futures* or *Gevent*. The AttributeProxy constructor internally makes some
    network calls which makes it *slow*. By using one of the *green modes* as
    green_mode you are allowing other python code to be executed in a cooperative way.

    :param full_attr_name: the full name of the attribute
    :type full_attr_name: str
    :param device_proxy: the :class:`~tango.DeviceProxy`
    :type device_proxy: DeviceProxy
    :param attr_name: attribute name for the given device proxy
    :type attr_name: str
    :param green_mode: determines the mode of execution of the device (including
                      the way it is created). Defaults to the current global
                      green_mode (check :func:`~tango.get_green_mode` and
                      :func:`~tango.set_green_mode`)
    :type green_mode: :obj:`~tango.GreenMode`
    :param wait: whether or not to wait for result. If green_mode
                 Ignored when green_mode is Synchronous (always waits).
    :type wait: bool
    :param timeout: The number of seconds to wait for the result.
                    If None, then there is no limit on the wait time.
                    Ignored when green_mode is Synchronous or wait is False.
    :type timeout: float
    :returns:
        if green_mode is Synchronous or wait is True:
            :class:`~tango.AttributeProxy`
        else if green_mode is Futures:
            :class:`concurrent.futures.Future`
        else if green_mode is Gevent:
            :class:`gevent.event.AsynchResult`
    :throws:
        * a *DevFailed* if green_mode is Synchronous or wait is True
          and there is an error creating the attribute.
        * a *concurrent.futures.TimeoutError* if green_mode is Futures,
          wait is False, timeout is not None and the time to create the attribute
          has expired.
        * a *gevent.timeout.Timeout* if green_mode is Gevent, wait is False,
          timeout is not None and the time to create the attribute has expired.

    New in PyTango 8.1.0
    """
    return AttributeProxy(*args, green_mode=green_mode)


def __AttributeProxy__get_property(
    self,
    propname: (
        str
        | DbDatum
        | DbData
        | list[str | bytes | DbDatum]
        | dict[str, DbDatum]
        | dict[str, list[str]]
        | dict[str, object]
    ),
    value=None,
) -> dict[str, list[str]]:
    """

    Get a (list) property(ies) for an attribute.

    :param propname: Can be one of the following: \n
                    1. :py:obj:`str` [in] - Single property data to be fetched. \n
                    2. :py:obj:`~tango.DbDatum` [in] - Single property data to be fetched. \n
                    3. :py:obj:`~tango.DbData` [in] - Several property data to be fetched. \n
                    4. :py:obj:`list`\\[:py:obj:`str` | :py:obj:`bytes`] [in] - Several property data to be fetched. \n
                    5. :py:obj:`list`\\[:py:obj:`tango.DbDatum`] [in] - Several property data to be fetched. \n
                    6. :py:obj:`dict`\\[:py:obj:`str`, :py:obj:`object`] [in] - Keys are property names
                       to be fetched (values are ignored). \n
                    7. :py:obj:`dict`\\[:py:obj:`str`, :obj:`tango.DbDatum`] [in] - Several `DbDatum.name` are
                       property names to be fetched (keys are ignored). \n

    :param value: Optional. For propname overloads with :py:obj:`str` and :py:obj:`list`\\[:py:obj:`str`] will be filed with the property values, if provided.
    :type value: :obj:`tango.DbData`, optional

    :returns: A :obj:`dict` object, which keys are the property names the value
              associated with each key being a sequence of strings being the
              property value.

    :throws:
        :py:obj:`TypeError`: Raised in case of propname has the wrong type. \n
        :py:obj:`tango.NonDbDevice`: Raised in case of a non-database device error. \n
        :py:obj:`tango.ConnectionFailed`: Raised on connection failure with the database. \n
        :py:obj:`tango.CommunicationFailed`: Raised on communication failure with the database. \n
        :py:obj:`tango.DevFailed`: Raised on a device failure from the database device.`

    .. versionadded:: 10.1.0: overloads with :obj:`dict` as propname parameter

    .. versionchanged:: 10.1.0: raises if propname has an invalid type instead of returning None
    """

    return get_property_from_db(self, propname, value)


def __AttributeProxy__put_property(
    self,
    value: (
        str
        | DbDatum
        | DbData
        | list[str | bytes | DbDatum]
        | dict[str, DbDatum]
        | dict[str, list[str]]
        | dict[str, object]
    ),
) -> None:
    """
    Insert or update a list of properties for this attribute.

    :param value: Can be one of the following: \n
                    1. :py:obj:`str` - Single property data to be inserted. \n
                    2. :py:obj:`~tango.DbDatum` - Single property data to be inserted. \n
                    3. :py:obj:`~tango.DbData` - Several property data to be inserted. \n
                    4. :py:obj:`list`\\[:py:obj:`str` | :py:obj:`bytes` | :py:obj:`~tango.DbDatum`] - Several property data to be inserted. \n
                    5. :py:obj:`dict`\\[:py:obj:`str`, :py:obj:`~tango.DbDatum`] -
                        DbDatum is property to be inserted (keys are ignored). \n
                    6. :py:obj:`dict`\\[:py:obj:`str`, :py:obj:`list`\\[:py:obj:`str`]] - Keys are property names,
                        and value has data to be inserted. \n
                    7. :py:obj:`dict`\\[:py:obj:`str`, :py:obj:`object`] - Keys are property names, and `str(obj)` is property value.

    :throws:
        :py:obj:`TypeError`: Raised in case of value has the wrong type. \n
        :py:obj:`tango.NonDbDevice`: Raised in case of a non-database device error. \n
        :py:obj:`tango.ConnectionFailed`: Raised on connection failure with the database. \n
        :py:obj:`tango.CommunicationFailed`: Raised on communication failure with the database. \n
        :py:obj:`tango.DevFailed`: Raised on a device failure from the database device.`
    """
    value = parameter_2_dbdata(value, "value")
    return self._put_property(value)


def __AttributeProxy__delete_property(
    self,
    value: (
        str
        | DbDatum
        | DbData
        | list[str | bytes | DbDatum]
        | dict[str, DbDatum]
        | dict[str, list[str]]
        | dict[str, object]
    ),
) -> None:
    """
    Delete a the given of properties for this attribute.
    :param value: Can be one of the following: \n
                    1. :py:obj:`str` [in] - Single property data to be deleted. \n
                    2. :py:obj:`~tango.DbDatum` [in] - Single property data to be deleted. \n
                    3. :py:obj:`~tango.DbData` [in] - Several property data to be deleted. \n
                    4. :py:obj:`list`\\[:py:obj:`str` | :py:obj:`bytes` | :py:obj:`~tango.DbDatum`] [in] - Several property data to be deleted. \n
                    5. :py:obj:`dict`\\[:py:obj:`str`, :py:obj:`object`] [in] - Keys are property names
                       to be deleted (values are ignored). \n
                    6. :py:obj:`dict`\\[:py:obj:`str`, :obj:`tango.DbDatum`] [in] - Several `DbDatum.name` are
                       property names to be deleted (keys are ignored). \n

    :throws:
        :py:obj:`TypeError`: Raised in case of value has the wrong type. \n
        :py:obj:`tango.NonDbDevice`: Raised in case of a non-database device error. \n
        :py:obj:`tango.ConnectionFailed`: Raised on connection failure with the database. \n
        :py:obj:`tango.CommunicationFailed`: Raised on communication failure with the database. \n
        :py:obj:`tango.DevFailed`: Raised on a device failure from the database device.`
    """

    new_value = parameter_2_dbdata(value, "value")
    return self._delete_property(new_value)


# It is easier to reimplement AttributeProxy in python using DeviceProxy than
# wrapping C++ AttributeProxy. However I still rely in the original
# AttributeProxy for the constructor (parsing strings if necessary) and some
# other things. With the _method_* functions defined later it is really easy.
# One reason to do it this way: get_device_proxy() will always return the
# same tango.DeviceProxy with this implementation. And then we can trust
# it's automatic event unsubscription to handle events.
class AttributeProxy:
    """
    AttributeProxy is the high level Tango object which provides the
    client with an easy-to-use interface to TANGO attributes.

    To create an AttributeProxy, a complete attribute name must be set
    in the object constructor.

    Example:
        att = AttributeProxy("tango/tangotest/1/long_scalar")

    Note: PyTango implementation of AttributeProxy is in part a
    python reimplementation of the AttributeProxy found on the C++ API.
    """

    @_trace_client
    def __init__(self, *args, green_mode=None):
        self.__initialized = False
        # If TestContext active, short TRL is replaced with fully-qualified
        # TRL, using test server's connection details.  Otherwise, left as-is.
        attr_name = args[0]
        new_attr_name = _get_device_fqtrl_if_necessary(attr_name)
        new_args = [new_attr_name] + list(args[1:])
        try:
            self.__attr_proxy = _AttributeProxy(*new_args)
        except DevFailed as orig_err:
            if new_attr_name != attr_name:
                # If attribute was not found, it could be an attempt to access a real
                # device with a short name while running TestContext.  I.e., we need
                # to use the short name so that the real TANGO_HOST will be tried.
                try:
                    self.__attr_proxy = _AttributeProxy(*args)
                except DevFailed as retry_exc:
                    Except.re_throw_exception(
                        retry_exc,
                        "PyAPI_AttributeProxyInitFailed",
                        f"Failed to create AttributeProxy "
                        f"(tried {new_attr_name!r} => {orig_err.args[0].reason}, and "
                        f"{attr_name!r} => {retry_exc.args[0].reason})",
                        "AttributeProxy.__init__",
                    )
            else:
                raise
        # get_device_proxy() returns a different python object each time
        # we don't want a different object, so we save the current one.
        self.__dev_proxy = dp = self.__attr_proxy.get_device_proxy()
        init_device_proxy(dp)
        dp.__dict__["_green_mode"] = (
            green_mode if green_mode is not None else get_green_mode()
        )
        self.__initialized = True

    def get_device_proxy(self):
        """
        get_device_proxy(self) -> DeviceProxy

                A method which returns the device associated to the attribute

            Parameters : None

            Return     : (DeviceProxy)
        """
        return self.__dev_proxy

    def name(self):
        """
        name(self) -> str

                Returns the attribute name

            Parameters : None
            Return     : (str) with the attribute name
        """
        if self.__initialized:
            name = self.__attr_proxy.name()
        else:
            name = "<Unknown: object was not fully initialized>"
        return name

    def __str__(self):
        return f"AttributeProxy({self.name()})"

    def __repr__(self):
        return f"AttributeProxy({self.name()})"


def _method_dev_and_name(dp_fn_name, doc=True):
    def __new_fn(self, *args, **kwds):
        return getattr(self._AttributeProxy__dev_proxy, dp_fn_name)(
            self.name(), *args, **kwds
        )

    if doc:
        __new_fn.__doc__ = (
            "This method is a simple way to do:\n"
            + "\tself.get_device_proxy()."
            + dp_fn_name
            + "(self.name(), ...)\n\n"
            + "For convenience, here is the documentation of DeviceProxy."
            + dp_fn_name
            + "(...):\n"
            + str(getattr(DeviceProxy, dp_fn_name).__doc__)
        )
    __new_fn.__name__ = dp_fn_name
    __new_fn.__qualname__ = f"AttributeProxy.{dp_fn_name}"
    return __new_fn


def _method_device(dp_fn_name, doc=True):
    def __new_fn(self, *args, **kwds):
        return getattr(self._AttributeProxy__dev_proxy, dp_fn_name)(*args, **kwds)

    if doc:
        __new_fn.__doc__ = (
            "This method is a simple way to do:\n"
            + "\tself.get_device_proxy()."
            + dp_fn_name
            + "(...)\n\n"
            + "For convenience, here is the documentation of DeviceProxy."
            + dp_fn_name
            + "(...):\n"
            + str(getattr(DeviceProxy, dp_fn_name).__doc__)
        )
    __new_fn.__name__ = dp_fn_name
    __new_fn.__qualname__ = f"AttributeProxy.{dp_fn_name}"
    return __new_fn


def _method_attribute(dp_fn_name, doc=True):
    def __new_fn(self, *args, **kwds):
        return getattr(self._AttributeProxy__attr_proxy, dp_fn_name)(*args, **kwds)

    if doc:
        __new_fn.__doc__ = getattr(_AttributeProxy, dp_fn_name).__doc__
    __new_fn.__name__ = dp_fn_name
    __new_fn.__qualname__ = f"AttributeProxy.{dp_fn_name}"
    return __new_fn


def attribute_proxy_init(doc=True):
    _AttributeProxy.get_property = __AttributeProxy__get_property
    _AttributeProxy.put_property = __AttributeProxy__put_property
    _AttributeProxy.delete_property = __AttributeProxy__delete_property

    # General methods
    # AttributeProxy.name                manually defined
    AttributeProxy.status = _method_device("status", doc=doc)
    AttributeProxy.state = _method_device("state", doc=doc)
    AttributeProxy.ping = _method_device("ping", doc=doc)
    AttributeProxy.get_transparency_reconnection = _method_device(
        "get_transparency_reconnection", doc=doc
    )
    AttributeProxy.set_transparency_reconnection = _method_device(
        "set_transparency_reconnection", doc=doc
    )

    # Property methods
    AttributeProxy.get_property = _trace_client(
        _method_attribute("get_property", doc=doc)
    )
    AttributeProxy.put_property = _trace_client(
        _method_attribute("put_property", doc=doc)
    )
    AttributeProxy.delete_property = _trace_client(
        _method_attribute("delete_property", doc=doc)
    )

    # Attribute methods
    AttributeProxy.get_config = _method_dev_and_name("get_attribute_config", doc=doc)
    AttributeProxy.set_config = _method_device("set_attribute_config", doc=doc)

    AttributeProxy.write = _method_dev_and_name("write_attribute", doc=doc)
    AttributeProxy.read = _method_dev_and_name("read_attribute", doc=doc)
    AttributeProxy.write_read = _method_dev_and_name("write_read_attribute", doc=doc)

    # History methods...
    AttributeProxy.history = _method_dev_and_name("attribute_history", doc=doc)

    # Polling administration methods
    AttributeProxy.poll = _method_dev_and_name("poll_attribute", doc=doc)
    AttributeProxy.get_poll_period = _method_dev_and_name(
        "get_attribute_poll_period", doc=doc
    )
    AttributeProxy.is_polled = _method_dev_and_name("is_attribute_polled", doc=doc)
    AttributeProxy.stop_poll = _method_dev_and_name("stop_poll_attribute", doc=doc)

    # Asynchronous methods
    AttributeProxy.read_asynch = _method_dev_and_name("read_attribute_asynch", doc=doc)
    AttributeProxy.read_reply = _method_device("read_attribute_reply", doc=doc)
    AttributeProxy.write_asynch = _method_dev_and_name(
        "write_attribute_asynch", doc=doc
    )
    AttributeProxy.write_reply = _method_device("write_attribute_reply", doc=doc)

    # Event methods
    AttributeProxy.subscribe_event = _method_dev_and_name("subscribe_event", doc=doc)
    AttributeProxy.unsubscribe_event = _method_device("unsubscribe_event", doc=doc)

    AttributeProxy.get_events = _method_device("get_events", doc=doc)
    AttributeProxy.event_queue_size = _method_device("event_queue_size", doc=doc)
    AttributeProxy.get_last_event_date = _method_device("get_last_event_date", doc=doc)
    AttributeProxy.is_event_queue_empty = _method_device(
        "is_event_queue_empty", doc=doc
    )



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""Test utilities"""

import enum
import time
from functools import wraps

import numpy as np

try:
    import numpy.typing as npt

    # in numpy 1.20 npt does not have npt.NDArray, so we cannot parce hits
    npt.NDArray
except (AttributeError, ImportError):
    npt = None

# Local imports
from tango import (
    DeviceProxy,
    DevState,
    GreenMode,
    AttrDataFormat,
    ExtractAs,
    DevFailed,
    DevVarLongStringArray,
    DevVarDoubleStringArray,
    SerialModel,
)
from tango.server import Device
from tango.test_context import (
    MultiDeviceTestContext,
    DeviceTestContext,
    get_server_port_via_pid,
)
from tango.utils import is_non_str_seq, FROM_TANGO_TO_NUMPY_TYPE
from tango import DeviceClass, LatestDeviceImpl, DevLong64, SCALAR, READ

# Conditional imports
try:
    import pytest
except ImportError:
    pytest = None

__all__ = [
    "MultiDeviceTestContext",
    "DeviceTestContext",
    "SimpleDevice",
    "ClassicAPISimpleDeviceImpl",
    "ClassicAPISimpleDeviceClass",
    "state",
    "command_typed_values",
    "attribute_typed_values",
    "dev_encoded_values",
    "server_green_mode",
    "attr_data_format",
    "convert_dtype_to_typing_hint",
    "assert_close",
    "general_decorator",
    "general_asyncio_decorator",
    "DEVICE_SERVER_ARGUMENTS",
    "wait_for_proxy",
    "wait_for_nodb_proxy_via_pid",
]

if npt:
    __all__ += [
        "command_numpy_typed_values",
        "attribute_numpy_typed_values",
        "attribute_wrong_numpy_typed",
    ]

UTF8_STRING = (
    r"""ăѣ𝔠ծềſģȟᎥ𝒋ǩľḿꞑȯ𝘱𝑞𝗋𝘴ȶ𝞄𝜈ψ𝒙𝘆𝚣1234567890!@#$%^&*()-_=+[{]};:'",<.>/?~𝘈Ḇ𝖢𝕯٤ḞԍНǏ𝙅ƘԸⲘ𝙉০Ρ𝗤Ɍ𝓢ȚЦ𝒱Ѡ𝓧ƳȤ"""
    r"""ả𝘢ѧᖯć𝗱ễ𝑓𝙜Ⴙ𝞲𝑗𝒌ļṃŉо𝞎𝒒ᵲꜱ𝙩ừ𝗏ŵ𝒙𝒚ź1234567890!@#$%^&*()-_=+[{]};:'",<.>/?~АḂⲤ𝗗𝖤𝗙ꞠꓧȊ𝐉𝜥ꓡ𝑀𝑵Ǭ𝙿𝑄Ŗ𝑆𝒯𝖴𝘝𝘞ꓫŸ𝜡"""
    r"""𝜶ƀ𝖼ḋếᵮℊ𝙝Ꭵ𝕛кιṃդⱺ𝓅𝘲𝕣𝖘ŧ𝑢ṽẉ𝘅ყž1234567890!@#$%^&*()-_=+[{]};:'",<.>/?~Ѧ𝙱ƇᗞΣℱԍҤ١𝔍К𝓛𝓜ƝȎ𝚸𝑄Ṛ𝓢ṮṺƲᏔꓫ𝚈𝚭"""
    r"""Ꮟçძ𝑒𝖿𝗀ḧ𝗂𝐣ҝɭḿ𝕟𝐨𝝔𝕢ṛ𝓼тú𝔳ẃ⤬𝝲𝗓1234567890!@#$%^&*()-_=+[{]};:'",<.>/?~𝖠Β𝒞𝘋𝙴𝓕ĢȞỈ𝕵ꓗʟ𝙼ℕ০𝚸𝗤ՀꓢṰǓⅤ𝔚Ⲭ𝑌𝙕𝘢𝕤"""
)

# char \x00 cannot be sent in a DevString. All other 1-255 chars can
ints = tuple(range(1, 256))
bytes_devstring = bytes(ints)
str_devstring = bytes_devstring.decode("latin-1")

# Test devices


class SimpleDevice(Device):
    def init_device(self):
        self.set_state(DevState.ON)


class ClassicAPISimpleDeviceImpl(LatestDeviceImpl):
    def __init__(self, cls, name):
        LatestDeviceImpl.__init__(self, cls, name)
        ClassicAPISimpleDeviceImpl.init_device(self)

    def init_device(self):
        self.get_device_properties(self.get_device_class())
        self.attr_attr1_read = 100

    def read_attr1(self, attr):
        attr.set_value(self.attr_attr1_read)


class ClassicAPISimpleDeviceClass(DeviceClass):
    attr_list = {"attr1": [[DevLong64, SCALAR, READ]]}


# Test enums


class GoodEnum(enum.IntEnum):
    START = 0
    MIDDLE = 1
    END = 2


class BadEnumNonZero(enum.IntEnum):
    START = 1
    MIDDLE = 2
    END = 3


class BadEnumSkipValues(enum.IntEnum):
    START = 0
    MIDDLE = 2
    END = 4


class BadEnumDuplicates(enum.IntEnum):
    START = 0
    MIDDLE = 1
    END = 1


# Helpers

# Note on Tango properties using the Tango File database:
# Tango file database cannot handle properties with '\n'. It doesn't
# handle '\' neither. And it cuts ASCII extended characters. That is
# why you will find that all property related tests are truncated to
# the first two values of the arrays below

GENERAL_TYPED_VALUES = {
    int: (1, 2, -65535, 23),
    float: (2.71, 3.14, -34.678e-10, 12.678e15),
    str: ("hey hey", "my my", bytes_devstring, str_devstring),
    bool: (False, True, True, False),
    (int,): (
        np.array([1, 2]),
        (1, 2, 3),
        [9, 8, 7],
        [-65535, 2224],
        [0, 0],
    ),
    (float,): (
        np.array([0.1, 0.2]),
        (0.1, 0.2, 0.3),
        [0.9, 0.8, 0.7],
        (-6.3232e-3, 1.234e4),
        [0.0, 12.56e12],
    ),
    (str,): (
        np.array(["foo", "bar"]),
        ("ab", "cd", "ef"),
        ["gh", "ij", "kl"],
        ("ab", "cd"),
        ["gh", "ij"],
        3 * [bytes_devstring],
        3 * [str_devstring],
    ),
    (bool,): (
        np.array([True, False]),
        (False, False, True),
        [True, False, False],
        (False, True),
        [True, False],
    ),
}

COMMAND_TYPED_VALUES = {
    DevVarLongStringArray: ([[1, 2, 3], ["foo", "bar", "hmm"]],),
    DevVarDoubleStringArray: ([[1.1, 2.2, 3.3], ["foo", "bar", "hmm"]],),
}

IMAGE_TYPED_VALUES = {
    ((int,),): (
        np.vstack((np.array([1, 2]), np.array([3, 4]))),
        ((1, 2, 3), (4, 5, 6), (7, 8, 9)),
        [[-65535, 2224, 23], [-6535, 224, 345], [-655, 24, 54]],
        ((1, 2, 3), (4, 5, 6)),
        [[-65535, 2224], [-65535, 2224]],
    ),
    ((float,),): (
        np.vstack((np.array([0.1, 0.2]), np.array([0.3, 0.4]))),
        ((0.1, 0.2, 0.3), (0.9, 0.8, 0.7), (0.5, 0.6, 0.7)),
        [[-6.3232e-3, 0.0], [0.0, 12.56e12], [23.4, 1.56e2]],
        ((0.1, 0.2, 0.3), (0.9, 0.8, 0.7)),
        [[-6.3232e-3, 0.0], [0.0, 12.56e12]],
    ),
    ((str,),): (
        np.vstack((np.array(["hi-hi", "ha-ha"]), np.array(["hu-hu", "yuhuu"]))),
        [["ab", "cd", "ef"], ["gh", "ij", "kl"], ["gh", "ij", "kl"]],
        (("ab", "cd", "ef"), ("gh", "ij", "kl"), ("gh", "ij", "kl")),
        [["ab", "cd"], ["ij", "kl"]],
        (("ab", "ef"), ("gh", "ij")),
        [3 * [bytes_devstring], 3 * [bytes_devstring]],
        [3 * [str_devstring], 3 * [str_devstring]],
    ),
    ((bool,),): (
        np.vstack((np.array([True, False]), np.array([False, True]))),
        [[False, False, True], [True, False, False], [True, False, False]],
        ((False, False, True), (True, False, False), (True, False, False)),
        [[False, True], [False, False]],
        ((False, True), (False, False)),
        [[False]],
        [[True]],
    ),
}

_numpy_hits_source = (
    ((np.bool_,), True, [True, False], [[True, False], [False, True]]),
    ((np.ubyte,), 1, [1, 2], [[1, 2], [3, 4]]),
    (
        (
            np.short,
            np.ushort,
            np.int32,
            np.uint32,
            np.int64,
            np.uint64,
        ),
        1,
        [1, 2],
        [[1, 2], [3, 4]],
    ),
    (
        (
            np.float64,
            np.float32,
        ),
        1.1,
        [1.1, 2.2],
        [[1.1, 2.2], [3.3, 4.4]],
    ),
)


if npt:
    NUMPY_GENERAL_TYPING_HINTS = []
    NUMPY_IMAGES_TYPING_HINTS = []
    for dtypes, scalar, spectrum, image in _numpy_hits_source:
        for dtype in dtypes:
            NUMPY_GENERAL_TYPING_HINTS.extend(
                [
                    [dtype, dtype, AttrDataFormat.SCALAR, scalar],
                    [dtype, npt.NDArray[dtype], AttrDataFormat.SPECTRUM, spectrum],
                ]
            )
            NUMPY_IMAGES_TYPING_HINTS.append(
                [dtype, npt.NDArray[dtype], AttrDataFormat.IMAGE, image]
            )

    WRONG_NUMPY_TYPING_HINTS = (  # dformat, max_x, max_y, value, error, match
        (None, None, None, None, RuntimeError, "AttrDataFormat has to be specified"),
        (
            AttrDataFormat.IMAGE,
            3,
            0,
            [[1, 2], [3, 4]],
            DevFailed,
            "Maximum y dim. wrongly defined",
        ),
        (
            AttrDataFormat.SPECTRUM,
            3,
            0,
            [[1, 2], [3, 4]],
            TypeError,
            "Expecting a integer type, but it is not",
        ),
    )

EXTRACT_AS = [
    (ExtractAs.Numpy, np.ndarray),
    (ExtractAs.Tuple, tuple),
    (ExtractAs.List, list),
    (ExtractAs.Bytes, bytes),
    (ExtractAs.ByteArray, bytearray),
    (ExtractAs.String, str),
]

BASE_TYPES = [float, int, str, bool]

# we also test a large dataset to force memory allocation from heap,
# to insure immediate segfault if we try to access data after dereferencing
LARGE_DATA_SIZE = 1 * 1024**2  # 1 Mb seems to be enough

DEV_ENCODED_DATA = {
    "str": UTF8_STRING,
    "bytes": UTF8_STRING.encode(),
    "bytearray": bytearray(UTF8_STRING.encode()),
}

# these sets to test Device Server input arguments

OS_SYSTEMS = ["linux", "win"]

#    os_system, in string, out arguments list, raised exception
DEVICE_SERVER_ARGUMENTS = (
    (
        ["linux", "win"],
        "MyDs instance --nodb --port 1234",
        ["MyDs", "instance", "-nodb", "-ORBendPoint", "giop:tcp:0.0.0.0:1234"],
    ),
    (
        ["linux", "win"],
        "MyDs -port 1234 -host myhost instance",
        ["MyDs", "instance", "-ORBendPoint", "giop:tcp:myhost:1234"],
    ),
    (
        ["linux", "win"],
        "MyDs instance --ORBendPoint giop:tcp::1234",
        ["MyDs", "instance", "-ORBendPoint", "giop:tcp::1234"],
    ),
    (
        ["linux", "win"],
        "MyDs instance -nodb -port 1000 -dlist a/b/c;d/e/f",
        [
            "MyDs",
            "instance",
            "-ORBendPoint",
            "giop:tcp:0.0.0.0:1000",
            "-nodb",
            "-dlist",
            "a/b/c;d/e/f",
        ],
    ),
    (
        ["linux", "win"],
        "MyDs instance -file a/b/c",
        ["MyDs", "instance", "-file=a/b/c"],
    ),
    ([], "MyDs instance -nodb", []),  # this test should always fail
    ([], "MyDs instance -dlist a/b/c;d/e/f", []),  # this test should always fail
    # the most complicated case: verbose
    (["linux", "win"], "MyDs instance -vvvv", ["MyDs", "instance", "-v4"]),
    (
        ["linux", "win"],
        "MyDs instance --verbose --verbose --verbose --verbose",
        ["MyDs", "instance", "-v4"],
    ),
    (["linux", "win"], "MyDs instance -v4", ["MyDs", "instance", "-v4"]),
    (["linux", "win"], "MyDs instance -v 4", ["MyDs", "instance", "-v4"]),
    # some options can be only in win, in linux should be error
    (
        ["win"],
        "MyDs instance -dbg -i -s -u",
        ["MyDs", "instance", "-dbg", "-i", "-s", "-u"],
    ),
    # variable ORB options
    (
        ["linux", "win"],
        "MyDs instance -ORBtest1 test1 --ORBtest2 test2",
        ["MyDs", "instance", "-ORBtest1", "test1", "-ORBtest2", "test2"],
    ),
    (
        ["linux", "win"],
        "MyDs ORBinstance -ORBtest myORBparam",
        ["MyDs", "ORBinstance", "-ORBtest", "myORBparam"],
    ),
    (
        ["linux", "win"],
        "MyDs instance -nodb -ORBendPoint giop:tcp:localhost:1234 -ORBendPointPublish giop:tcp:myhost.local:2345",
        [
            "MyDs",
            "instance",
            "-nodb",
            "-ORBendPoint",
            "giop:tcp:localhost:1234",
            "-ORBendPointPublish",
            "giop:tcp:myhost.local:2345",
        ],
    ),
    (
        [],
        "MyDs instance -ORBtest1 test1 --orbinvalid value",
        [],
    ),  # lowercase "orb" should fail
)


def convert_dtype_to_typing_hint(dtype):
    check_x_dim, check_y_dim = False, False
    if type(dtype) is tuple:
        dtype = dtype[0]
        check_x_dim = True
        if type(dtype) is tuple:
            dtype = dtype[0]
            check_y_dim = True
            tuple_hint = tuple[
                tuple[dtype, dtype, dtype],
                tuple[dtype, dtype, dtype],
                tuple[dtype, dtype, dtype],
                tuple[dtype, dtype, dtype],
            ]
            list_hint = list[list[dtype]]
        else:
            tuple_hint = tuple[dtype, dtype, dtype]
            list_hint = list[dtype]
    elif dtype == DevVarLongStringArray:
        tuple_hint = tuple[tuple[int], tuple[str]]
        list_hint = list[list[int], list[str]]
    elif dtype == DevVarDoubleStringArray:
        tuple_hint = tuple[tuple[float], tuple[str]]
        list_hint = list[list[float], list[str]]
    else:
        tuple_hint = dtype
        list_hint = dtype

    return tuple_hint, list_hint, check_x_dim, check_y_dim


def general_decorator(function=None):
    if function:

        @wraps(function)
        def _wrapper(*args, **kwargs):
            return function(*args, **kwargs)

        return _wrapper
    else:
        return general_decorator


def general_asyncio_decorator(function=None):
    if function:

        @wraps(function)
        async def _wrapper(*args, **kwargs):
            return await function(*args, **kwargs)

        return _wrapper
    else:
        return general_asyncio_decorator


def repr_type(x):
    if isinstance(x, (list, tuple)):
        return f"({repr_type(x[0])},)"
    elif x == DevVarLongStringArray:
        return "DevVarLongStringArray"
    elif x == DevVarDoubleStringArray:
        return "DevVarDoubleStringArray"
    return f"{x.__name__}"


def repr_numpy_type(dtype, dformat):
    if dformat == AttrDataFormat.SCALAR:
        return f"{dtype.__name__}, SCALAR"
    elif dformat == AttrDataFormat.SPECTRUM:
        return f"np.NDarray[{dtype.__name__}], SPECTRUM"
    else:
        return f"np.n2array[{dtype.__name__}], IMAGE"


# helpers to test enums
def check_attr_type(read_attr, attr_data_format, desired_type):
    if attr_data_format == AttrDataFormat.SCALAR:
        assert isinstance(read_attr, desired_type)
    elif attr_data_format == AttrDataFormat.SPECTRUM:
        for state in read_attr:
            assert isinstance(state, desired_type)
    else:
        for state in read_attr:
            for stat in state:
                assert isinstance(stat, desired_type)


def assert_value_label(read_attr, value, label):
    assert read_attr.value == value
    assert read_attr.name == label


def check_read_attr(read_attr, attr_data_format, value, label):
    if attr_data_format == AttrDataFormat.SCALAR:
        assert_value_label(read_attr, value, label)
    elif attr_data_format == AttrDataFormat.SPECTRUM:
        for val in read_attr:
            assert_value_label(val, value, label)
    else:
        for val in read_attr:
            for v in val:
                assert_value_label(v, value, label)


def make_nd_value(value, attr_data_format):
    if attr_data_format == AttrDataFormat.SCALAR:
        return value
    elif attr_data_format == AttrDataFormat.SPECTRUM:
        return (value,)
    else:
        return ((value,),)


# Numpy helpers
if pytest:

    def __assert_all_types(a, b):
        if isinstance(a, str):
            assert a == b
            return
        elif isinstance(a, dict):
            for k, v in a.items():
                assert k in b
                assert_close(v, b[k])
            return
        elif isinstance(a, (np.bool_, bool)) or isinstance(b, (np.bool_, bool)):
            assert a == b
            return
        try:
            assert a == pytest.approx(b)
        except (ValueError, TypeError):
            np.testing.assert_allclose(a, b)

    def assert_close(a, b):
        if is_non_str_seq(a):
            assert len(a) == len(b)
            for _a, _b in zip(a, b):
                assert_close(_a, _b)
        else:
            __assert_all_types(a, b)

    def __convert_value(value):
        if isinstance(value, bytes):
            return value.decode("latin-1")
        return value

    def create_result(dtype, value):
        if isinstance(dtype, (list, tuple)):
            dtype = dtype[0]
            return [create_result(dtype, v) for v in value]
        elif dtype == DevVarLongStringArray:
            return [create_result(dtype, v) for v, dtype in zip(value, [int, str])]
        elif dtype == DevVarDoubleStringArray:
            return [create_result(dtype, v) for v, dtype in zip(value, [float, str])]

        return __convert_value(value)

    def convert_to_type(value, attr_type, expected_type):
        if expected_type in [tuple, list]:
            return expected_type(value)
        elif expected_type == np.ndarray:
            return np.array(value, dtype=FROM_TANGO_TO_NUMPY_TYPE[attr_type])
        elif expected_type in [bytes, bytearray, str]:
            value = np.array(value, dtype=FROM_TANGO_TO_NUMPY_TYPE[attr_type]).tobytes()
            if expected_type is bytearray:
                return bytearray(value)
            elif expected_type is str:
                return "".join([chr(b) for b in value])
            return value
        else:
            pytest.xfail("Unknown extract_as type")

    @pytest.fixture(params=list(DevState.values.values()), ids=str)
    def state(request):
        return request.param

    @pytest.fixture(
        params=list(GENERAL_TYPED_VALUES.items()), ids=lambda x: repr_type(x[0])
    )
    def general_typed_values(request):
        dtype, values = request.param
        expected = lambda v: create_result(dtype, v)
        return dtype, values, expected

    @pytest.fixture(
        params=list({**GENERAL_TYPED_VALUES, **COMMAND_TYPED_VALUES}.items()),
        ids=lambda x: repr_type(x[0]),
    )
    def command_typed_values(request):
        dtype, values = request.param
        expected = lambda v: create_result(dtype, v)
        return dtype, values, expected

    @pytest.fixture(
        params=list({**GENERAL_TYPED_VALUES, **IMAGE_TYPED_VALUES}.items()),
        ids=lambda x: repr_type(x[0]),
    )
    def attribute_typed_values(request):
        dtype, values = request.param
        expected = lambda v: create_result(dtype, v)
        return dtype, values, expected

    if npt:

        @pytest.fixture(
            params=NUMPY_GENERAL_TYPING_HINTS, ids=lambda x: repr_numpy_type(x[0], x[2])
        )
        def command_numpy_typed_values(request):
            dtype, type_hint, dformat, values = request.param
            expected = lambda v: create_result(dtype, v)
            return type_hint, dformat, values, expected

        @pytest.fixture(
            params=NUMPY_GENERAL_TYPING_HINTS + NUMPY_IMAGES_TYPING_HINTS,
            ids=lambda x: repr_numpy_type(x[0], x[2]),
        )
        def attribute_numpy_typed_values(request):
            dtype, type_hint, dformat, values = request.param
            expected = lambda v: create_result(dtype, v)
            return type_hint, dformat, values, expected

        @pytest.fixture(
            params=WRONG_NUMPY_TYPING_HINTS, ids=lambda x: f"{x[-2].__name__}: {x[-1]}"
        )
        def attribute_wrong_numpy_typed(request):
            return request.param

    else:

        @pytest.fixture
        def command_numpy_typed_values(request):
            raise RuntimeError(
                f"Numpy typing supported only for Numpy >= 1.20, "
                f"while current version is {np.version.version}"
            )

        @pytest.fixture
        def attribute_numpy_typed_values(request):
            raise RuntimeError(
                f"Numpy typing supported only for Numpy >= 1.20, "
                f"while current version is {np.version.version}"
            )

        @pytest.fixture
        def attribute_wrong_numpy_typed(request):
            raise RuntimeError(
                f"Numpy typing supported only for Numpy >= 1.20, "
                f"while current version is {np.version.version}"
            )

    @pytest.fixture(
        params=list(DEV_ENCODED_DATA.items()),
        ids=list(DEV_ENCODED_DATA.keys()),
    )
    def dev_encoded_values(request):
        return request.param

    @pytest.fixture(
        params=EXTRACT_AS, ids=[f"extract_as.{req_type}" for req_type, _ in EXTRACT_AS]
    )
    def extract_as(request):
        requested_type, expected_type = request.param
        return requested_type, expected_type

    @pytest.fixture(params=BASE_TYPES)
    def base_type(request):
        return request.param

    @pytest.fixture(params=list(GreenMode.values.values()), ids=str)
    def green_mode(request):
        return request.param

    @pytest.fixture(
        params=[GreenMode.Synchronous, GreenMode.Asyncio, GreenMode.Gevent], ids=str
    )
    def server_green_mode(request):
        return request.param

    @pytest.fixture(
        params=[AttrDataFormat.SCALAR, AttrDataFormat.SPECTRUM, AttrDataFormat.IMAGE],
        ids=str,
    )
    def attr_data_format(request):
        return request.param

    @pytest.fixture(
        params=[
            SerialModel.BY_DEVICE,
            SerialModel.BY_CLASS,
            SerialModel.BY_PROCESS,
            SerialModel.NO_SYNC,
        ],
        ids=str,
    )
    def server_serial_model(request):
        return request.param


def wait_for_proxy(
    dev_name: str,
    proxy_class: type[DeviceProxy] = DeviceProxy,
    retries: int = 600,
    delay: float = 0.02,
) -> DeviceProxy:
    """Create a new DeviceProxy, retrying until it is responding.

    The DeviceProxy will try to ping the device and read the state at least once.

    :param dev_name: device name for DeviceProxy connection string
    :param proxy_class: Type of DeviceProxy class to instantiate (could be synchronous,
                        asyncio, or gevent versions of the class).  If using asyncio,
                        then provide: ``partial(tango.asyncio.DeviceProxy, wait=True)``.
    :param retries: number of times to retry attempts, optional
    :param delay: time to wait (seconds) between retries, optional

    :returns: DeviceProxy object created to access the specified device

    :raises RuntimeError: If the device is not responding before the timeout

    .. versionadded:: 10.1.0
    """
    proxy = None
    last_error = ""
    count = 0
    while count < retries:
        try:
            proxy = proxy_class(dev_name)
            break
        except DevFailed as exc:
            last_error = str(exc)
            time.sleep(delay)
        count += 1
    if proxy is not None:
        while count < retries:
            try:
                proxy.ping()
                proxy.state()
                return proxy
            except DevFailed as exc:
                last_error = str(exc)
                time.sleep(delay)
            count += 1
    raise RuntimeError(
        f"Device at {dev_name} did not respond within {count * delay:.1f} sec!\n"
        f"Last error: {last_error}."
    )


def wait_for_nodb_proxy_via_pid(
    pid: int,
    host: str,
    dev_name: str,
    proxy_class: type[DeviceProxy],
    retries: int = 600,
    delay: float = 0.02,
) -> DeviceProxy:
    """Create new DeviceProxy, with retrying until it is ready and responding.

    The PID is used to get the process information and probe it to find the correct
    TCP port number to connect to.  Creation of the DeviceProxy is retried until
    the device is responding, or times out.  The DeviceProxy will ping the
    device and read the state.

    :param pid: operating system process identifier
    :param host: hostname/IP that device server is listening on.
    :param dev_name: device name for DeviceProxy connection string
    :param proxy_class: Type of DeviceProxy class to instantiate (could be synchronous,
                        asyncio, or gevent versions of the class).  If using asyncio,
                        then provide: ``partial(tango.asyncio.DeviceProxy, wait=True)``.
    :param retries: number of times to retry attempts, optional
    :param delay: time to wait (seconds) between retries, optional

    :returns: DeviceProxy object created to access the specified device

    :raises RuntimeError: If the GIOP port couldn't be identified, or proxy timed out

    .. seealso::
        :func:`tango.test_context.get_server_port_via_pid`

    .. versionadded:: 10.1.0
    """
    t0 = time.time()
    port = get_server_port_via_pid(pid, host, retries, delay)

    timeout = retries * delay
    elapsed = time.time() - t0
    time_left = timeout - elapsed
    retries_left = max(1, int(time_left / delay))
    trl = f"tango://{host}:{port}/{dev_name}#dbase=no"
    return wait_for_proxy(trl, proxy_class, retries_left, delay)



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
"""
Example used for debugging during 4th Tango Kernel Webinar.

https://www.tango-controls.org/community/news/2021/06/10/4th-tango-kernel-webinar-pytango/
"""

from tango import AttrQuality, AttrWriteType, GreenMode
from tango.server import Device, attribute
from tango.test_utils import DeviceTestContext, assert_close


def test_read_attribute():
    class TestDevice(Device):
        _voltage = 0.0

        @attribute(dtype=float, access=AttrWriteType.READ_WRITE)
        def voltage(self):
            return self._voltage

        @voltage.write
        def voltage(self, value):
            self._voltage = value

    with DeviceTestContext(TestDevice) as proxy:
        # low-level API read
        low_level_api_reading = proxy.read_attribute("voltage")
        assert_close(low_level_api_reading.value, 0.0)
        assert low_level_api_reading.quality is AttrQuality.ATTR_VALID

        # high-level API read
        high_level_api_read_value = proxy.voltage
        assert_close(high_level_api_read_value, 0.0)

        # low-level API write
        value = 24.0
        proxy.write_attribute("voltage", value)
        high_level_api_read_value = proxy.voltage
        assert_close(high_level_api_read_value, value)

        # high-level API write
        value = 48.0
        proxy.voltage = value
        high_level_api_read_value = proxy.voltage
        assert_close(high_level_api_read_value, value)


def test_read_attribute_full():
    class TestDevice(Device):
        green_mode = GreenMode.Synchronous

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._voltage = 0.0

        def init_device(self):
            super().init_device()
            print("init_device!")

        def read_attr_hardware(self, attr_list):
            print(f"read_attr_hardware for {str(attr_list)}")

        def write_attr_hardware(self, attr_list):
            print(f"write_attr_hardware for {str(attr_list)}")

        def always_executed_hook(self):
            print("always_executed_hook")

        @attribute(dtype=float, access=AttrWriteType.READ_WRITE)
        def voltage(self):
            return self._voltage

        @voltage.write
        def voltage(self, value):
            self._voltage = value

        def is_voltage_allowed(self, req_type):
            return True

        @attribute(dtype=float, access=AttrWriteType.READ)
        def current(self):
            return self._voltage / 2.0

    with DeviceTestContext(TestDevice, timeout=600, process=False) as proxy:
        # low-level API read single
        low_level_api_reading = proxy.read_attribute("voltage")
        assert_close(low_level_api_reading.value, 0.0)
        assert low_level_api_reading.quality is AttrQuality.ATTR_VALID

        # low-level API read multiple
        low_level_api_readings = proxy.read_attributes(["voltage", "current"])
        assert_close(low_level_api_readings[0].value, 0.0)
        assert low_level_api_readings[0].quality is AttrQuality.ATTR_VALID

        # high-level API read
        high_level_api_read_value = proxy.voltage
        assert_close(high_level_api_read_value, 0.0)

        # low-level API write
        value = 24.0
        proxy.write_attribute("voltage", value)
        high_level_api_read_value = proxy.voltage
        assert_close(high_level_api_read_value, value)

        # high-level API write
        value = 48.0
        proxy.voltage = value
        high_level_api_read_value = proxy.voltage
        assert_close(high_level_api_read_value, value)



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

import time

from tango import DeviceProxy
from tango.test_context import DeviceTestContext
from tango.server import Device, attribute


class TestDevice(Device):
    _value = 0.0

    @attribute
    def double_scalar(self) -> float:
        self._value += 1.0
        return self._value

    @double_scalar.write
    def double_scalar(self, value: float) -> None:
        self._value = value


def eval_telemetry_overhead(
    device_name, num_reads_per_iteration=100, num_iterations=200
):
    time_per_iteration = []
    dp = DeviceProxy(device_name)

    for i in range(num_iterations):
        start = time.perf_counter()
        for _ in range(num_reads_per_iteration):
            _ = dp.read_attribute("double_scalar")
        end = time.perf_counter()
        duration_ms = (end - start) * 1000.0
        print(
            f"  {i:3d} Total execution time for {num_reads_per_iteration} iterations: {duration_ms:.3f} milliseconds."
        )
        tpi = duration_ms / num_reads_per_iteration
        print(f"  {i:3d} Average execution time per iteration: {tpi:.3f} milliseconds")
        time_per_iteration.append(tpi)

    average_tpi = sum(time_per_iteration) / num_iterations
    square_sum = 0.0
    for tpi in time_per_iteration:
        square_sum += (tpi - average_tpi) * (tpi - average_tpi)

    rms_tpi = (square_sum / num_iterations) ** 0.5

    print(f"Average execution time: {average_tpi:.3f} milliseconds.")
    print(f"Standard deviation of execution times: {rms_tpi:.3f} milliseconds.")


def benchmark_python_client_and_python_server():
    with DeviceTestContext(TestDevice, device_name="test/device/1", process=True):
        print("Python server running, and runing Python client benchmark now...")
        eval_telemetry_overhead(
            "test/device/1", num_reads_per_iteration=100, num_iterations=100
        )


def benchmark_cpp_client_and_python_server():
    context = DeviceTestContext(
        TestDevice, device_name="test/device/1", process=True, port=44555
    )
    with context:
        print(
            f"Python server running - run C++ client benchmark now for {context.get_device_access()}"
        )
        time.sleep(100)


def benchmark_python_client_and_cpp_server():
    eval_telemetry_overhead(
        "sys/tg_test/1", num_reads_per_iteration=100, num_iterations=100
    )


if __name__ == "__main__":
    # Run one of the following:
    benchmark_python_client_and_python_server()
    # benchmark_cpp_client_and_python_server()
    # benchmark_python_client_and_cpp_server()



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
"""Provides PyTango as a module for backward compatibility."""

# Imports
import sys
import tango
import pkgutil


def alias_package(package, alias, extra_modules={}):
    """Alias a python package properly.

    It ensures that modules are not duplicated by trying
    to import and alias all the submodules recursively.
    """
    path = package.__path__
    alias_prefix = alias + "."
    prefix = package.__name__ + "."
    # Alias all importable modules recursively
    for _, name, _ in pkgutil.walk_packages(path, prefix):
        # Skip databaseds backends
        if name.startswith("tango.databaseds.db_access."):
            continue
        try:
            if name not in sys.modules:
                __import__(name)
        except ImportError:
            continue
        alias_name = name.replace(prefix, alias_prefix)
        sys.modules[alias_name] = sys.modules[name]
    # Alias extra modules
    for key, value in extra_modules.items():
        name = prefix + value
        if name not in sys.modules:
            __import__(name)
        if not hasattr(package, key):
            setattr(package, key, sys.modules[name])
        sys.modules[alias_prefix + key] = sys.modules[name]
    # Alias root module
    sys.modules[alias] = sys.modules[package.__name__]


# Do not flood pytango users console with warnings yet
# warnings.warn('PyTango module is deprecated, import tango instead.')

# Alias tango package
alias_package(
    package=tango,
    alias=__name__,
    extra_modules={"_PyTango": "_tango", "constants": "constants"},
)



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
This is the main PyTango package file.
Documentation for this package can be found online:

http://pytango.readthedocs.io
"""


__all__ = (
    "AccessControlType",
    "ApiUtil",
    "ArchiveEventInfo",
    "ArchiveEventProp",
    "ArgType",
    "AsynCall",
    "AsynReplyNotArrived",
    "AttReqType",
    "Attr",
    "AttrConfEventData",
    "AttrData",
    "AttrDataFormat",
    "AttrMemorizedType",
    "AttrProperty",
    "AttrQuality",
    "AttrReadEvent",
    "AttrSerialModel",
    "AttrWriteType",
    "AttrWrittenEvent",
    "Attribute",
    "AttributeAlarm",
    "AttributeAlarmInfo",
    "AttributeConfig",
    "AttributeConfig_2",
    "AttributeConfig_3",
    "AttributeConfig_5",
    "AttributeDimension",
    "AttributeEventInfo",
    "AttributeInfo",
    "AttributeInfoEx",
    "AttributeInfoList",
    "AttributeInfoListEx",
    "AttributeProxy",
    "ChangeEventInfo",
    "ChangeEventProp",
    "DevIntrChangeEventData",
    "ClientAddr",
    "CmdArgType",
    "CmdDoneEvent",
    "CommandInfo",
    "CommandInfoList",
    "CommunicationFailed",
    "Connection",
    "ConnectionFailed",
    "ConstDevString",
    "DServer",
    "DataReadyEventData",
    "Database",
    "DbData",
    "DbDatum",
    "DbDevExportInfo",
    "DbDevExportInfos",
    "DbDevImportInfo",
    "DbDevImportInfos",
    "DbDevFullInfo",
    "DbDevInfo",
    "DbDevInfos",
    "DbHistory",
    "DbHistoryList",
    "DbServerInfo",
    "DbServerData",
    "DebugIt",
    "DevBoolean",
    "DevCommandInfo",
    "DevDouble",
    "DevEncoded",
    "DevError",
    "DevFailed",
    "DevFloat",
    "DevLong",
    "DevLong64",
    "DevShort",
    "DevSource",
    "DevString",
    "DevUChar",
    "DevULong",
    "DevULong64",
    "DevUShort",
    "DevEnum",
    "DevState",
    "DevVarBooleanArray",
    "DevVarCharArray",
    "DevVarDoubleArray",
    "DevVarDoubleStringArray",
    "DevVarFloatArray",
    "DevVarLong64Array",
    "DevVarLongArray",
    "DevVarLongStringArray",
    "DevVarShortArray",
    "DevVarStringArray",
    "DevVarULong64Array",
    "DevVarULongArray",
    "DevVarUShortArray",
    "DevVoid",
    "DeviceAttribute",
    "DeviceAttributeConfig",
    "DeviceAttributeHistory",
    "DeviceClass",
    "DeviceData",
    "DeviceDataList",
    "DeviceDataHistory",
    "DeviceDataHistoryList",
    "DeviceImpl",
    "DeviceInfo",
    "DeviceProxy",
    "DeviceUnlocked",
    "Device_2Impl",
    "Device_3Impl",
    "Device_4Impl",
    "Device_5Impl",
    "Device_6Impl",
    "DispLevel",
    "EncodedAttribute",
    "EnsureOmniThread",
    "ErrSeverity",
    "ErrorIt",
    "EventData",
    "EventProperties",
    "EventReason",
    "EventSubMode",
    "EventSystemFailed",
    "EventType",
    "Except",
    "ExtractAs",
    "FMT_UNKNOWN",
    "FatalIt",
    "GreenMode",
    "Group",
    "GroupAttrReply",
    "GroupAttrReplyList",
    "GroupCmdReply",
    "GroupCmdReplyList",
    "GroupReply",
    "GroupReplyList",
    "IMAGE",
    "ImageAttr",
    "InfoIt",
    "KeepAliveCmdCode",
    "Level",
    "LockCmdCode",
    "LockerInfo",
    "LockerLanguage",
    "LogIt",
    "LogLevel",
    "LogTarget",
    "Logger",
    "Logging",
    "MessBoxType",
    "MultiAttribute",
    "MultiAttrProp",
    "MultiClassAttribute",
    "NamedDevFailed",
    "NamedDevFailedList",
    "NonDbDevice",
    "NonSupportedFeature",
    "NotAllowed",
    "PeriodicEventInfo",
    "PeriodicEventProp",
    "PollCmdCode",
    "PollDevice",
    "PollObjType",
    "PyTangoUserWarning",
    "READ",
    "READ_WITH_WRITE",
    "READ_WRITE",
    "Release",
    "SCALAR",
    "SPECTRUM",
    "SerialModel",
    "SpectrumAttr",
    "StdDoubleVector",
    "StdLongVector",
    "StdNamedDevFailedVector",
    "StdStringVector",
    "SubDevDiag",
    "TangoStream",
    "TimeVal",
    "TimedAttrData",
    "TimedCmdData",
    "UserDefaultAttrProp",
    "Util",
    "WAttribute",
    "WRITE",
    "WarnIt",
    "WrongData",
    "WrongNameSyntax",
    "__version__",
    "__version_description__",
    "__version_info__",
    "__version_long__",
    "__version_number__",
    "alarm_flags",
    "asyn_req_type",
    "cb_sub_model",
    "class_factory",
    "class_list",
    "constants",
    "constructed_class",
    "cpp_class_list",
    "delete_class_list",
    "get_class",
    "get_classes",
    "get_constructed_class",
    "get_constructed_classes",
    "get_cpp_class",
    "get_cpp_classes",
    "raise_asynch_exception",
    "AutoTangoMonitor",
    "AutoTangoAllowThreads",
    "LatestDeviceImpl",
    "Interceptors",
    "get_attribute_proxy",
    "requires_tango",
    "requires_pytango",
    "set_green_mode",
    "get_green_mode",
    "get_device_proxy",
    "is_scalar_type",
    "is_array_type",
    "is_omni_thread",
    "is_numerical_type",
    "is_int_type",
    "is_float_type",
    "is_bool_type",
    "is_str_type",
    "obj_2_str",
    "str_2_obj",
    "seqStr_2_obj",
)

__docformat__ = "restructuredtext"

from ._tango import (
    AccessControlType,
    ApiUtil,
    ArchiveEventInfo,
    AsynCall,
    AsynReplyNotArrived,
    AttReqType,
    Attr,
    AttrConfEventData,
    AttrDataFormat,
    AttrMemorizedType,
    AttrProperty,
    AttrQuality,
    AttrReadEvent,
    AttrSerialModel,
    AttrWriteType,
    AttrWrittenEvent,
    Attribute,
    AttributeAlarm,
    AttributeAlarmInfo,
    AttributeConfig,
    AttributeConfig_2,
    AttributeConfig_3,
    AttributeConfig_5,
    AttributeDimension,
    AttributeEventInfo,
    AttributeInfo,
    AttributeInfoEx,
    AttributeInfoList,
    AttributeInfoListEx,
    ArchiveEventProp,
    ChangeEventInfo,
    ChangeEventProp,
    ClientAddr,
    CmdArgType,
    DevIntrChangeEventData,
    CmdDoneEvent,
    CommandInfo,
    CommandInfoList,
    CommunicationFailed,
    Connection,
    ConnectionFailed,
    ConstDevString,
    DServer,
    DataReadyEventData,
    Database,
    DbData,
    DbDatum,
    DbDevExportInfo,
    DbDevExportInfos,
    DbDevImportInfo,
    DbDevImportInfos,
    DbDevFullInfo,
    DbDevInfo,
    DbDevInfos,
    DbHistory,
    DbHistoryList,
    DbServerInfo,
    DbServerData,
    DevBoolean,
    DevCommandInfo,
    DevDouble,
    DevEncoded,
    DevError,
    DevFailed,
    DevFloat,
    DevLong,
    DevLong64,
    DevShort,
    DevSource,
    DevString,
    DevUChar,
    DevULong,
    DevULong64,
    DevUShort,
    DevEnum,
    DevState,
    DevVarBooleanArray,
    DevVarCharArray,
    DevVarDoubleArray,
    DevVarDoubleStringArray,
    DevVarFloatArray,
    DevVarLong64Array,
    DevVarLongArray,
    DevVarLongStringArray,
    DevVarShortArray,
    DevVarStringArray,
    DevVarULong64Array,
    DevVarULongArray,
    DevVarUShortArray,
    DevVoid,
    DeviceAttribute,
    DeviceAttributeConfig,
    DeviceAttributeHistory,
    DeviceData,
    DeviceDataList,
    DeviceDataHistory,
    DeviceDataHistoryList,
    DeviceImpl,
    DeviceInfo,
    DeviceProxy,
    DeviceUnlocked,
    Device_2Impl,
    Device_3Impl,
    Device_4Impl,
    Device_5Impl,
    Device_6Impl,
    DispLevel,
    EncodedAttribute,
    EnsureOmniThread,
    ErrSeverity,
    EventData,
    EventSystemFailed,
    EventProperties,
    EventReason,
    EventSubMode,
    EventType,
    Except,
    ExtractAs,
    GreenMode,
    FMT_UNKNOWN,
    GroupAttrReply,
    GroupAttrReplyList,
    GroupCmdReply,
    GroupCmdReplyList,
    GroupReply,
    GroupReplyList,
    IMAGE,
    ImageAttr,
    KeepAliveCmdCode,
    Level,
    LockCmdCode,
    LockerInfo,
    LockerLanguage,
    LogLevel,
    LogTarget,
    Logger,
    Logging,
    MessBoxType,
    MultiAttribute,
    MultiClassAttribute,
    NamedDevFailed,
    NamedDevFailedList,
    NonDbDevice,
    NonSupportedFeature,
    NotAllowed,
    PeriodicEventInfo,
    PeriodicEventProp,
    PollCmdCode,
    PollDevice,
    PollObjType,
    READ,
    READ_WITH_WRITE,
    READ_WRITE,
    SCALAR,
    SPECTRUM,
    SerialModel,
    SpectrumAttr,
    StdDoubleVector,
    StdLongVector,
    StdNamedDevFailedVector,
    StdStringVector,
    SubDevDiag,
    TimeVal,
    UserDefaultAttrProp,
    WAttribute,
    WRITE,
    WrongData,
    WrongNameSyntax,
    alarm_flags,
    asyn_req_type,
    cb_sub_model,
    constants,
    raise_asynch_exception,
    Interceptors,
    AutoTangoMonitor,
    AutoTangoAllowThreads,
    is_omni_thread,
)

__auto_die_callbacks_holder = {}


# Aliases

ArgType = CmdArgType

# Release

from .release import Release

__author__ = Release.author_lines
__version_info__ = Release.version_info
__version_short__ = Release.version
__version__ = Release.version_long
__version_long__ = Release.version_long
__version_number__ = Release.version_number
__version_description__ = Release.version_description
__doc__ = Release.long_description

# Pytango imports

from .attr_data import AttrData

from .log4tango import TangoStream, LogIt, DebugIt, InfoIt, WarnIt, ErrorIt, FatalIt

from .device_server import (
    MultiAttrProp,
    LatestDeviceImpl,
)

from .attribute_proxy import AttributeProxy, get_attribute_proxy

from .group import Group

from .pyutil import Util

from .device_class import DeviceClass

from .globals import (
    get_class,
    get_classes,
    get_cpp_class,
    get_cpp_classes,
    get_constructed_class,
    get_constructed_classes,
    class_factory,
    delete_class_list,
    class_list,
    cpp_class_list,
    constructed_class,
)

from .utils import (
    requires_pytango,
    requires_tango,
    is_scalar_type,
    is_array_type,
    is_numerical_type,
    is_int_type,
    is_float_type,
    is_bool_type,
    is_str_type,
    obj_2_str,
    str_2_obj,
    seqStr_2_obj,
    PyTangoUserWarning,
    _RemovedClass,
)

from .green import set_green_mode, get_green_mode

from .device_proxy import get_device_proxy

from .pyutil import TimedAttrData, TimedCmdData

# PyTango initialization

from .pytango_init import init as __init

PipeWriteType = _RemovedClass(
    "PipeWriteType Enum was removed from PyTango starting from version 10.1.0. Pipes are not supported anymore. "
    "Please update your code!"
)

__init()
requires_tango("10.1.0", software_name="PyTango")



#!/usr/bin/env python3
# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
Trivial power supply device with no external connection or behaviour.

Extends ps0b.py, changing the voltage attribute to push events
directly, rather than using polling.
This is a poor example since changes in the voltage attribute
will only get pushed at the end of calibration, rather than
whenever they change "significantly".
"""

import random
from time import sleep
from tango.server import Device, attribute, command


class PowerSupply(Device):
    def init_device(self):
        self.set_change_event("voltage", True, False)
        self.set_archive_event("voltage", True, False)

    @attribute(dtype=float)
    def voltage(self):
        return self._get_voltage()

    def _get_voltage(self):
        noise = -0.05 + 0.1 * random.random()
        return 1.5 + noise

    @command
    def calibrate(self):
        sleep(0.1)
        value = self._get_voltage()
        self.push_change_event("voltage", value)
        self.push_archive_event("voltage", value)


if __name__ == "__main__":
    PowerSupply.run_server()
